
/*****************************************************************************/
/*                                                                           */
/*  THE NRCONV NURSE ROSTERING TO XHSTT CONVERTER                            */
/*  COPYRIGHT (C) 2016, Jeffrey H. Kingston                                  */
/*                                                                           */
/*  Jeffrey H. Kingston (jeff@it.usyd.edu.au)                                */
/*  School of Information Technologies                                       */
/*  The University of Sydney 2006                                            */
/*  AUSTRALIA                                                                */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU General Public License as published by     */
/*  the Free Software Foundation; either Version 3, or (at your option)      */
/*  any later version.                                                       */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston MA 02111-1307 USA   */
/*                                                                           */
/*  FILE:         nrc_demand.c                                               */
/*  MODULE:       Demands                                                    */
/*                                                                           */
/*****************************************************************************/
#include "nrc_interns.h"

#define DEBUG1 0
#define DEBUG2 0
#define DEBUG3 0


/*****************************************************************************/
/*                                                                           */
/*  NRC_PENALTY_GROUP                                                        */
/*                                                                           */
/*  A group of workers who should *not* be penalized by p;                   */
/*  these become the preferred workers in a prefer resources constraint.     */
/*                                                                           */
/*****************************************************************************/

typedef struct nrc_penalty_group_rec {
  NRC_PENALTY		penalty;
  NRC_WORKER_SET	worker_set;
} *NRC_PENALTY_GROUP;

typedef HA_ARRAY(NRC_PENALTY_GROUP) ARRAY_NRC_PENALTY_GROUP;


/*****************************************************************************/
/*                                                                           */
/*  NRC_DEMAND - a demand for one nurse                                      */
/*                                                                           */
/*****************************************************************************/

struct nrc_demand_rec {
  NRC_INSTANCE			instance;
  char				*name;
  NRC_PENALTY			not_assigned_penalty;
  NRC_PENALTY			assigned_penalty;
  NRC_WORKER_SET		preferred_ws;
  NRC_PENALTY			not_preferred_penalty;
  NRC_SHIFT			prev_shift;
  int				prev_shift_index;
  int				max_shift_index;
  NRC_SHIFT_SET			shift_set;
  ARRAY_NRC_PENALTY_GROUP	penalty_groups;
};


/*****************************************************************************/
/*                                                                           */
/*  Submodule "penalty groups (private)"                                     */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  NRC_PENALTY_GROUP NrcPenaltyGroupMake(NRC_DEMAND d, NRC_PENALTY p)       */
/*                                                                           */
/*  Make, add, and return a penalty weight object with these attributes.     */
/*                                                                           */
/*****************************************************************************/

static NRC_PENALTY_GROUP NrcPenaltyGroupMake(NRC_DEMAND d, NRC_PENALTY p)
{
  NRC_PENALTY_GROUP res;  char *name;  HA_ARENA a;
  a = NrcInstanceArena(d->instance);
  if( DEBUG2 )
    fprintf(stderr, "[ NrcPenaltyGroupMake(%s, %s)\n", d->name,
      NrcPenaltyShow(p, a));
  HaMake(res, a);
  res->penalty = p;
  name = HnStringMake(a, "%s-Or-UnpreferredNot=%s",
    NrcWorkerSetName(d->preferred_ws), NrcPenaltyId(p));
  if( !NrcInstanceRetrieveWorkerSet(d->instance, name, &res->worker_set) )
    res->worker_set = NrcWorkerSetMake(d->instance, name);
  HaArrayAddLast(d->penalty_groups, res);
  if( DEBUG2 )
    fprintf(stderr, "] res = %p, res->worker_set = %p\n",
      (void *) res, (void *) res->worker_set);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  NRC_PENALTY_GROUP NrcPenaltyGroupDefaultMake(NRC_DEMAND d, NRC_PENALTY p)*/
/*                                                                           */
/*  Make, add, and return a default penalty weight object with these         */
/*  attributes.                                                              */
/*                                                                           */
/*****************************************************************************/

static NRC_PENALTY_GROUP NrcPenaltyGroupDefaultMake(NRC_DEMAND d, NRC_PENALTY p)
{
  NRC_PENALTY_GROUP res;  HA_ARENA a;
  a = NrcInstanceArena(d->instance);
  if( DEBUG2 )
    fprintf(stderr, "[ NrcPenaltyGroupDefaultMake(%s, %s)\n", d->name,
      NrcPenaltyShow(p, a));
  HaMake(res, a);
  res->penalty = p;
  res->worker_set = d->preferred_ws;
  HaArrayAddLast(d->penalty_groups, res);
  if( DEBUG2 )
    fprintf(stderr, "] res = %p, res->worker_set = %p\n",
      (void *) res, (void *) res->worker_set);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  bool NrcPenaltyGroupRetrieve(NRC_DEMAND d, NRC_PENALTY p,                */
/*    NRC_PENALTY_GROUP *pg)                                                 */
/*                                                                           */
/*  Retrieve a penalty group with this penalty from d, if present.           */
/*                                                                           */
/*****************************************************************************/

static bool NrcPenaltyGroupRetrieve(NRC_DEMAND d, NRC_PENALTY p,
  NRC_PENALTY_GROUP *pg)
{
  int i;  NRC_PENALTY_GROUP pg2;
  HaArrayForEach(d->penalty_groups, pg2, i)
    if( NrcPenaltyEqual(pg2->penalty, p) )
    {
      *pg = pg2;
      return true;
    }
  *pg = NULL;
  return false;
}


/*****************************************************************************/
/*                                                                           */
/*  void NrcPenaltyGroupDebug(NRC_PENALTY_GROUP pg, int indent, FILE *fp)    */
/*                                                                           */
/*  Debug print of pg onto fp with the given indent.                         */
/*                                                                           */
/*****************************************************************************/

static void NrcPenaltyGroupDebug(NRC_PENALTY_GROUP pg, int indent, FILE *fp)
{
  int i;  NRC_WORKER w;  HA_ARENA a;
  a = NrcInstanceArena(NrcWorkerSetInstance(pg->worker_set));
  if( indent >= 0 )
  {
    fprintf(fp, "%*s[ PenaltyGroup(%s)", indent, "",
      NrcPenaltyShow(pg->penalty, a));
    for( i = 0;  i < NrcWorkerSetWorkerCount(pg->worker_set);  i++ )
    {
      w = NrcWorkerSetWorker(pg->worker_set, i);
      if( i % 5 == 0 )
	fprintf(fp, "%s\n%*s", i == 0 ? "" : ",", indent + 4, "");
      else
	fprintf(fp, ", ");
      fprintf(fp, "%s", NrcWorkerConvertedName(w));
    }
    fprintf(fp, "\n%*s]\n", indent, "");
  }
  else
    fprintf(fp, "PenaltyGroup(%s, %s)", NrcPenaltyShow(pg->penalty, a),
      NrcWorkerSetName(pg->worker_set));
}


/*****************************************************************************/
/*                                                                           */
/*  void NrcDemandBuildPenaltyGroups(NRC_DEMAND d)                           */
/*                                                                           */
/*  Build the penalty groups for d.                                          */
/*                                                                           */
/*****************************************************************************/

static void NrcDemandBuildPenaltyGroups(NRC_DEMAND d)
{
  bool ok;  int i, j;  NRC_WORKER w, w2;  NRC_PENALTY_GROUP pg;
  NRC_PENALTY p;

  if( DEBUG1 )
    fprintf(stderr, "[ NrcDemandBuildPenaltyGroups(%s)\n", d->name);

  if( NrcInstanceUniqueSkillPenalty(d->instance, &p) )
  {
    /* all workers have the same skill penalty, so we need one penalty group */
    if( p == NULL )
      p = d->not_preferred_penalty;
    pg = NrcPenaltyGroupDefaultMake(d, p);
  }
  else
  {
    /* build one penalty group for each distinct non-zero penalty */
    for( i = 0;  i < NrcInstanceStaffingWorkerCount(d->instance);  i++ )
    {
      /* get one worker w and its skill penalty, possibly the default */
      w = NrcInstanceStaffingWorker(d->instance, i);
      p = NrcWorkerSkillPenalty(w);
      if( p == NULL )
	p = d->not_preferred_penalty;
      if( DEBUG1 )
      {
	HA_ARENA a = NrcInstanceArena(d->instance);
	fprintf(stderr, "  worker %s: %s\n", NrcWorkerConvertedName(w),
	  NrcPenaltyShow(p, a));
      }

      /* make sure there is a penalty group with this penalty, if non-zero */
      if( NrcPenaltyWeight(p) > 0 && !NrcPenaltyGroupRetrieve(d, p, &pg) )
	pg = NrcPenaltyGroupMake(d, p);
    }

    /* add each worker to each penalty group as required */
    for( i = 0;  i < NrcInstanceStaffingWorkerCount(d->instance);  i++ )
    {
      /* get one worker w and its skill penalty, possibly the default */
      w = NrcInstanceStaffingWorker(d->instance, i);
      p = NrcWorkerSkillPenalty(w);
      if( p == NULL )
	p = d->not_preferred_penalty;

      /* add w to each penalty group it qualifies for, either because it */
      /* is in preferred_ws, or because its penalty differs from the group's */
      ok = NrcWorkerSetRetrieveWorker(d->preferred_ws, NrcWorkerName(w), &w2);
      HaArrayForEach(d->penalty_groups, pg, j)
	if( ok || (pg != NULL && !NrcPenaltyEqual(p, pg->penalty)) )
	  if( !NrcWorkerSetContainsWorker(pg->worker_set, w) )
	    NrcWorkerSetAddWorker(pg->worker_set, w);
    }
  }
  if( DEBUG1 )
  {
    HaArrayForEach(d->penalty_groups, pg, j)
      NrcPenaltyGroupDebug(pg, 2, stderr);
    fprintf(stderr, "]\n");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "construction and query"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  char *NrcDemandMakeName(NRC_PENALTY not_assigned_penalty,                */
/*    NRC_PENALTY assigned_penalty, NRC_WORKER_SET preferred_ws,             */
/*    NRC_PENALTY not_preferred_penalty)                                     */
/*                                                                           */
/*  Return the name of a demand with these attributes.                       */
/*                                                                           */
/*****************************************************************************/

static char *NrcDemandMakeName(NRC_PENALTY not_assigned_penalty,
  NRC_PENALTY assigned_penalty, NRC_WORKER_SET preferred_ws,
  NRC_PENALTY not_preferred_penalty, HA_ARENA a)
{
  char buff1[100], buff3[100], *buff2, *name, *p;  char *items[3];
  int item_count;
  item_count = 0;

  /* sort out the part of the name relating to being not assigned */
  if( NrcPenaltyWeight(not_assigned_penalty) > 0 )
  {
    sprintf(buff1, "A=%s", NrcPenaltyId(not_assigned_penalty));
    items[item_count++] = buff1;
  }

  /* sort out the part of the name relating to preferred workers */
  if( preferred_ws != NULL )
  {
    p = NrcWorkerSetName(preferred_ws);
    if( strstr(p, "Skill-") == p )
      name = &p[strlen("Skill-")];
    else
      name = p;
    buff2 = HnStringMake(a, "P-%s=%s", name,
      NrcPenaltyId(not_preferred_penalty));
    items[item_count++] = buff2;
  }

  /* sort out the part of the name relating to being assigned */
  if( NrcPenaltyWeight(assigned_penalty) > 0 )
  {
    sprintf(buff3, "U=%s", NrcPenaltyId(assigned_penalty));
    items[item_count++] = buff3;
  }

  /* return the name */
  switch( item_count )
  {
    case 0:	return HnStringMake(a, "X");
    case 1:	return HnStringMake(a, "%s", items[0]);
    case 2:	return HnStringMake(a, "%s:%s", items[0], items[1]);
    case 3:	return HnStringMake(a, "%s:%s:%s", items[0], items[1],items[2]);

    default:

      HnAbort("NrcDemandMakeName internal error");
      return NULL;  /* keep compiler happy */
  }
}


/*****************************************************************************/
/*                                                                           */
/*  NRC_DEMAND NrcDemandMake(NRC_INSTANCE ins, bool not_assigned_reqd,       */
/*    int not_assigned_weight, bool assigned_reqd, int assigned_weight,      */
/*    NRC_WORKER_SET preferred_ws, bool not_preferred_reqd,                  */
/*    int not_preferred_weight)                                              */
/*                                                                           */
/*  Either make a new role or retrieve an existing role with these           */
/*  attributes.                                                              */
/*                                                                           */
/*****************************************************************************/

NRC_DEMAND NrcDemandMake(NRC_INSTANCE ins,
  NRC_PENALTY not_assigned_penalty, NRC_PENALTY assigned_penalty,
  NRC_WORKER_SET preferred_ws, NRC_PENALTY not_preferred_penalty)
{
  char *name;  NRC_DEMAND res;  HA_ARENA a;

  /* make sure all penalties are non-NULL */
  HnAssert(not_assigned_penalty != NULL,
    "NrcDemandMake: NULL not_assigned_penalty");
  HnAssert(assigned_penalty != NULL, "NrcDemandMake: NULL assigned_penalty");
  HnAssert(not_preferred_penalty != NULL || preferred_ws == NULL,
    "NrcDemandMake: NULL not_preferred_penalty");

  /* normalize the parameters */
  if( preferred_ws == NULL )
    not_preferred_penalty = NrcInstanceZeroPenalty(ins);

  /* make the name */
  a = NrcInstanceArena(ins);
  name = NrcDemandMakeName(not_assigned_penalty, assigned_penalty,
    preferred_ws, not_preferred_penalty, a);

  /* retrieve or make the demand object, and return it */
  if( !NrcInstanceRetrieveDemand(ins, name, &res) )
  {
    HaMake(res, a);
    res->instance = ins;
    res->name = name;
    res->not_assigned_penalty = not_assigned_penalty;
    res->assigned_penalty = assigned_penalty;
    res->preferred_ws = preferred_ws;
    res->not_preferred_penalty = not_preferred_penalty;
    res->prev_shift = NULL;
    res->prev_shift_index = 0;
    res->max_shift_index = 0;
    res->shift_set = NrcShiftSetMake(ins);
    HaArrayInit(res->penalty_groups, a);
    NrcInstanceAddDemand(ins, res);
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  NRC_INSTANCE NrcDemandInstance(NRC_DEMAND d)                             */
/*                                                                           */
/*  Return the instance attribute of d.                                      */
/*                                                                           */
/*****************************************************************************/

NRC_INSTANCE NrcDemandInstance(NRC_DEMAND d)
{
  return d->instance;
}


/*****************************************************************************/
/*                                                                           */
/*  NRC_PENALTY NrcDemandNotAssignedPenalty(NRC_DEMAND d)                    */
/*                                                                           */
/*  Return the not_assigned_penalty attribute of d.                          */
/*                                                                           */
/*****************************************************************************/

NRC_PENALTY NrcDemandNotAssignedPenalty(NRC_DEMAND d)
{
  return d->not_assigned_penalty;
}


/*****************************************************************************/
/*                                                                           */
/*  NRC_PENALTY NrcDemandAssignedPenalty(NRC_DEMAND d)                       */
/*                                                                           */
/*  Return the assigned_penalty attribute of d.                              */
/*                                                                           */
/*****************************************************************************/

NRC_PENALTY NrcDemandAssignedPenalty(NRC_DEMAND d)
{
  return d->assigned_penalty;
}


/*****************************************************************************/
/*                                                                           */
/*  NRC_WORKER_SET NrcDemandPreferredWorkerSet(NRC_DEMAND d)                 */
/*                                                                           */
/*  Return the preferred_ws attribute of d.                                  */
/*                                                                           */
/*****************************************************************************/

NRC_WORKER_SET NrcDemandPreferredWorkerSet(NRC_DEMAND d)
{
  return d->preferred_ws;
}


/*****************************************************************************/
/*                                                                           */
/*  NRC_PENALTY NrcDemandNotPreferredPenalty(NRC_DEMAND d)                   */
/*                                                                           */
/*  Return the not_preferred_reqd attribute of d.                            */
/*                                                                           */
/*****************************************************************************/

NRC_PENALTY NrcDemandNotPreferredPenalty(NRC_DEMAND d)
{
  return d->not_preferred_penalty;
}


/*****************************************************************************/
/*                                                                           */
/*  char *NrcDemandName(NRC_DEMAND d)                                        */
/*                                                                           */
/*  Return the name attribute of d.                                          */
/*                                                                           */
/*****************************************************************************/

char *NrcDemandName(NRC_DEMAND d)
{
  return d->name;
}


/*****************************************************************************/
/*                                                                           */
/*  bool NrcDemandCanUseDemands(NRC_DEMAND d)                                */
/*                                                                           */
/*  Return true if d is convertible to a demand from a demand constraint.    */
/*                                                                           */
/*****************************************************************************/

/* ***
bool NrcDemandCanUseDemands(NRC_DEMAND d)
{
  return NrcPenaltyWeight(d->not_assigned_penalty) == 0 &&
    NrcPenaltyWeight(d->assigned_penalty) == 0;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  char *NrcDemandRoleMake(NRC_DEMAND d, int shift_index)                   */
/*                                                                           */
/*  Make a role name for demand d with index.                                */
/*                                                                           */
/*****************************************************************************/

static char *NrcDemandRoleMake(NRC_DEMAND d, int index)
{
  HA_ARENA a = NrcInstanceArena(d->instance);
  return HnStringMake(a, "%s:%d", d->name, index);
}


/*****************************************************************************/
/*                                                                           */
/*  void NrcDemandConvertToEventResource(NRC_DEMAND d, NRC_SHIFT s,          */
/*    KHE_EVENT e)                                                           */
/*                                                                           */
/*  Convert d into an event resource for s, whose KHE event is e.            */
/*                                                                           */
/*  Implementation note.  This function assumes that all calls for a         */
/*  given shift s occur together, not separated by calls for other           */
/*  shifts, as is indeed the case.                                           */
/*                                                                           */
/*****************************************************************************/

void NrcDemandConvertToEventResource(NRC_DEMAND d, NRC_SHIFT s,
  NRC_WORKER preasst, KHE_EVENT e)
{
  KHE_RESOURCE_TYPE rt;  KHE_EVENT_RESOURCE er;  char *role;
  KHE_RESOURCE r;

  /* sort out prev_shift, prev_shift_index, and max_shift_index */
  if( d->prev_shift == s )
    d->prev_shift_index++;
  else
  {
    d->prev_shift = s;
    d->prev_shift_index = 1;
    NrcShiftSetAddShift(d->shift_set, s);
  }
  if( d->max_shift_index < d->prev_shift_index )
    d->max_shift_index = d->prev_shift_index;

  /* generate the event resource */
  rt = KheInstanceResourceType(KheEventInstance(e), 0);
  r = (preasst == NULL ? NULL : NrcWorkerResource(preasst));
  role = NrcDemandRoleMake(d, d->prev_shift_index);
  if( !KheEventResourceMake(e, rt, r, role, KheEventWorkload(e), &er) )
    HnAbort("NrcDemandConvertToEventResource internal error: %s", role);
}


/*****************************************************************************/
/*                                                                           */
/*  void NrcDemandConvertToEventResourceConstraints(NRC_DEMAND d,            */
/*    KHE_INSTANCE ins)                                                      */
/*                                                                           */
/*  Convert d into all corresponding event resources constraints.            */
/*                                                                           */
/*****************************************************************************/

void NrcDemandConvertToEventResourceConstraints(NRC_DEMAND d, KHE_INSTANCE ins)
{
  KHE_EVENT e;  KHE_ASSIGN_RESOURCE_CONSTRAINT karc;  HA_ARENA a;
  int index, i, staffing_count;  KHE_EVENT_GROUP eg;  NRC_PENALTY_GROUP pg;
  KHE_PREFER_RESOURCES_CONSTRAINT kprc;  char *role, *id, *name;
  a = NrcInstanceArena(d->instance);
  id = HnStringMake(a, "EG:%s", d->name);
  eg = NrcShiftSetEventGroup(d->shift_set, id, ins);
  for( index = 1;  index <= d->max_shift_index;  index++ )
  {
    /* not assigned */
    role = NrcDemandRoleMake(d, index);
    if( NrcPenaltyWeight(d->not_assigned_penalty) > 0 )
    {
      id = HnStringMake(a, "ARC:%s", role);
      name = "Assign resource";
      if( !KheAssignResourceConstraintMake(ins, id, name,
	  NrcPenaltyConvert(d->not_assigned_penalty), role, &karc) )
	HnAbort("cannot add %s constraint", id);
      KheAssignResourceConstraintAddEventGroup(karc, eg);
    }

    /* assigned */
    if( NrcPenaltyWeight(d->assigned_penalty) > 0 )
    {
      id = HnStringMake(a, "PRC:%s", role);
      name = "Do not assign resource";
      if( !KhePreferResourcesConstraintMake(ins, id, name,
          NrcPenaltyConvert(d->assigned_penalty), role, &kprc) )
	HnAbort("cannot add %s constraint", id);
      if( !KhePreferResourcesConstraintAddResourceGroup(kprc,
	  NrcInstanceEmptyResourceGroup(d->instance)) )
	HnAbort("cannot add empty rg to %s constraint", id);
      if( !KhePreferResourcesConstraintAddEventGroup(kprc, eg, &e) )
	HnAbort("cannot add all_eg to %s constraint (problem is %s)",
	  id, KheEventId(e));
    }

    /* not preferred */
    if( d->preferred_ws != NULL )
    {
      if( index == 1 )
	NrcDemandBuildPenaltyGroups(d);
      staffing_count = NrcInstanceStaffingWorkerCount(d->instance);
      HaArrayForEach(d->penalty_groups, pg, i)
	if( NrcWorkerSetWorkerCount(pg->worker_set) < staffing_count )
	{
	  id = HnStringMake(a, "PRC:%s:Not=%s", role,NrcPenaltyId(pg->penalty));
	  name = HnStringMake(a, "Prefer %s resource",
	    NrcWorkerSetName(pg->worker_set));
	  if( !KhePreferResourcesConstraintMake(ins, id, name,
	      NrcPenaltyConvert(pg->penalty), role, &kprc) )
	    HnAbort("cannot add %s constraint", id);
	  if( !KhePreferResourcesConstraintAddResourceGroup(kprc,
		NrcWorkerSetResourceGroup(pg->worker_set, ins)) )
	    HnAbort("cannot add rg to %s constraint", id);
	  if( !KhePreferResourcesConstraintAddEventGroup(kprc, eg, &e) )
	    HnAbort("cannot add all_eg to %s constraint (problem is %s)",
	      id, KheEventId(e));
	}
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  int NrcDemandTypedCmp(NRC_DEMAND d1, NRC_DEMAND d2)                      */
/*                                                                           */
/*  Comparison function for sorting an array of demands by name.             */
/*                                                                           */
/*****************************************************************************/

int NrcDemandTypedCmp(NRC_DEMAND d1, NRC_DEMAND d2)
{
  return strcmp(d1->name, d2->name);
}


/*****************************************************************************/
/*                                                                           */
/*  int NrcDemandCmp(const void *t1, const void *t2)                         */
/*                                                                           */
/*  Untyped demand comparison function.                                      */
/*                                                                           */
/*****************************************************************************/


int NrcDemandCmp(const void *t1, const void *t2)
{
  NRC_DEMAND d1 = * (NRC_DEMAND *) t1;
  NRC_DEMAND d2 = * (NRC_DEMAND *) t2;
  return NrcDemandTypedCmp(d1, d2);
}


/*****************************************************************************/
/*                                                                           */
/*  void NrcDemandDebug(NRC_DEMAND d, int indent, FILE *fp)                  */
/*                                                                           */
/*  Debug print of d onto fp.                                                */
/*                                                                           */
/*****************************************************************************/

void NrcDemandDebug(NRC_DEMAND d, int indent, FILE *fp)
{
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  fprintf(fp, "%s", d->name);
  if( indent >= 0 )
    fprintf(fp, "\n");
}
