
/*****************************************************************************/
/*                                                                           */
/*  THE KHE HIGH SCHOOL TIMETABLING ENGINE                                   */
/*  COPYRIGHT (C) 2010 Jeffrey H. Kingston                                   */
/*                                                                           */
/*  Jeffrey H. Kingston (jeff@it.usyd.edu.au)                                */
/*  School of Information Technologies                                       */
/*  The University of Sydney 2006                                            */
/*  AUSTRALIA                                                                */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU General Public License as published by     */
/*  the Free Software Foundation; either Version 3, or (at your option)      */
/*  any later version.                                                       */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston MA 02111-1307 USA   */
/*                                                                           */
/*  FILE:         khe_sr_frame.c                                             */
/*  DESCRIPTION:  Resource frames                                            */
/*                                                                           */
/*****************************************************************************/
#include "khe_solvers.h"
#include <limits.h>
#define min(a, b) ((a) < (b) ? (a) : (b))
#define max(a, b) ((a) > (b) ? (a) : (b))


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KheFrameTaskSet"                                              */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  bool KheFrameTaskSet(KHE_FRAME frame, KHE_RESOURCE r, KHE_TASK_SET *ts) */
/*                                                                           */
/*  Return a new task set containing the tasks assigned to r in frame.       */
/*  Return true if none of these tasks are preassigned.                      */
/*                                                                           */
/*****************************************************************************/

bool KheFrameTaskSet(KHE_FRAME frame, KHE_RESOURCE r, KHE_TASK_SET *ts)
{
  KHE_RESOURCE_TIMETABLE_MONITOR rtm;  int i, j, k, pos, count;
  KHE_TIME_GROUP tg;  KHE_TIME t;  KHE_TASK task;
  KHE_SOLN soln;  KHE_RESOURCE r2;  bool res;
  soln = KheFrameSoln(frame);
  res = true;
  *ts = KheTaskSetMake(soln);
  rtm = KheResourceTimetableMonitor(soln, r);
  for( i = 0;  i < KheFrameTimeGroupCount(frame);  i++ )
  {
    tg = KheFrameTimeGroup(frame, i);
    for( j = 0;  j < KheTimeGroupTimeCount(tg);  j++ )
    {
      t = KheTimeGroupTime(tg, j);
      count = KheResourceTimetableMonitorTimeTaskCount(rtm, t);
      for( k = 0;  k < count;  k++ )
      {
	task = KheResourceTimetableMonitorTimeTask(rtm, t, k);
	task = KheTaskProperRoot(task);
	if( task != NULL && !KheTaskSetContainsTask(*ts, task, &pos) )
	{
	  KheTaskSetAddTask(*ts, task);
	  if( KheTaskIsPreassigned(task, &r2) )
	    res = false;
	}
      }
    }
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "frames - visiting"                                            */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheFrameSetVisitNum(KHE_FRAME frame, KHE_RESOURCE r, int num)       */
/*                                                                           */
/*  Call KheTaskSetVisitNum on each task of (frame, r).                      */
/*                                                                           */
/*****************************************************************************/

void KheFrameSetVisitNum(KHE_FRAME frame, KHE_RESOURCE r, int num)
{
  KHE_RESOURCE_TIMETABLE_MONITOR rtm;  int i, j, k, count;  KHE_SOLN soln;
  KHE_TIME_GROUP tg;  KHE_TIME t;  KHE_TASK task;
  soln = KheFrameSoln(frame);
  rtm = KheResourceTimetableMonitor(soln, r);
  for( i = 0;  i < KheFrameTimeGroupCount(frame);  i++ )
  {
    tg = KheFrameTimeGroup(frame, i);
    for( j = 0;  j < KheTimeGroupTimeCount(tg);  j++ )
    {
      t = KheTimeGroupTime(tg, j);
      count = KheResourceTimetableMonitorTimeTaskCount(rtm, t);
      for( k = 0;  k < count;  k++ )
      {
	task = KheResourceTimetableMonitorTimeTask(rtm, t, k);
	task = KheTaskProperRoot(task);
	if( task != NULL )
	  KheTaskSetVisitNum(task, num);
      }
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  int KheFrameVisitNum(KHE_FRAME frame, KHE_RESOURCE r)                    */
/*                                                                           */
/*  Call KheTaskVisitNum on any task of (frame, r).                          */
/*                                                                           */
/*****************************************************************************/

int KheFrameVisitNum(KHE_FRAME frame, KHE_RESOURCE r)
{
  KHE_RESOURCE_TIMETABLE_MONITOR rtm;  int i, j, k, count;  KHE_SOLN soln;
  KHE_TIME_GROUP tg;  KHE_TIME t;  KHE_TASK task;
  soln = KheFrameSoln(frame);
  rtm = KheResourceTimetableMonitor(soln, r);
  for( i = 0;  i < KheFrameTimeGroupCount(frame);  i++ )
  {
    tg = KheFrameTimeGroup(frame, i);
    for( j = 0;  j < KheTimeGroupTimeCount(tg);  j++ )
    {
      t = KheTimeGroupTime(tg, j);
      count = KheResourceTimetableMonitorTimeTaskCount(rtm, t);
      for( k = 0;  k < count;  k++ )
      {
	task = KheResourceTimetableMonitorTimeTask(rtm, t, k);
	task = KheTaskProperRoot(task);
	if( task != NULL )
	  return KheTaskVisitNum(task);
      }
    }
  }
  HnAbort("KheFrameVisitNum:  resource frame has no tasks");
  return 0;  /* keep compiler happy */
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheFrameVisited(KHE_FRAME frame, KHE_RESOURCE r, int slack)         */
/*                                                                           */
/*  Return true if any call to KheTaskVisited(task, slack) for task in       */
/*  (frame, r) returns true.                                                 */
/*                                                                           */
/*****************************************************************************/

bool KheFrameVisited(KHE_FRAME frame, KHE_RESOURCE r, int slack)
{
  KHE_RESOURCE_TIMETABLE_MONITOR rtm;  int i, j, k, count;  KHE_SOLN soln;
  KHE_TIME_GROUP tg;  KHE_TIME t;  KHE_TASK task;
  soln = KheFrameSoln(frame);
  rtm = KheResourceTimetableMonitor(soln, r);
  for( i = 0;  i < KheFrameTimeGroupCount(frame);  i++ )
  {
    tg = KheFrameTimeGroup(frame, i);
    for( j = 0;  j < KheTimeGroupTimeCount(tg);  j++ )
    {
      t = KheTimeGroupTime(tg, j);
      count = KheResourceTimetableMonitorTimeTaskCount(rtm, t);
      for( k = 0;  k < count;  k++ )
      {
	task = KheResourceTimetableMonitorTimeTask(rtm, t, k);
	task = KheTaskProperRoot(task);
	if( task != NULL && KheTaskVisited(task, slack) )
	  return true;
      }
    }
  }
  return false;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheFrameVisit(KHE_FRAME frame, KHE_RESOURCE r)                      */
/*                                                                           */
/*  Mark the tasks of (frame, r) visited.                                    */
/*                                                                           */
/*****************************************************************************/

void KheFrameVisit(KHE_FRAME frame, KHE_RESOURCE r)
{
  KHE_RESOURCE_TIMETABLE_MONITOR rtm;  int i, j, k, count;  KHE_SOLN soln;
  KHE_TIME_GROUP tg;  KHE_TIME t;  KHE_TASK task;
  soln = KheFrameSoln(frame);
  rtm = KheResourceTimetableMonitor(soln, r);
  for( i = 0;  i < KheFrameTimeGroupCount(frame);  i++ )
  {
    tg = KheFrameTimeGroup(frame, i);
    for( j = 0;  j < KheTimeGroupTimeCount(tg);  j++ )
    {
      t = KheTimeGroupTime(tg, j);
      count = KheResourceTimetableMonitorTimeTaskCount(rtm, t);
      for( k = 0;  k < count;  k++ )
      {
	task = KheResourceTimetableMonitorTimeTask(rtm, t, k);
	task = KheTaskProperRoot(task);
	if( task != NULL )
	  KheTaskVisit(task);
      }
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheFrameUnVisit(KHE_FRAME frame, KHE_RESOURCE r)                    */
/*                                                                           */
/*  Mark the tasks of (frame, r) unvisited.                                  */
/*                                                                           */
/*****************************************************************************/

void KheFrameUnVisit(KHE_FRAME frame, KHE_RESOURCE r)
{
  KHE_RESOURCE_TIMETABLE_MONITOR rtm;  int i, j, k, count;  KHE_SOLN soln;
  KHE_TIME_GROUP tg;  KHE_TIME t;  KHE_TASK task;
  soln = KheFrameSoln(frame);
  rtm = KheResourceTimetableMonitor(soln, r);
  for( i = 0;  i < KheFrameTimeGroupCount(frame);  i++ )
  {
    tg = KheFrameTimeGroup(frame, i);
    for( j = 0;  j < KheTimeGroupTimeCount(tg);  j++ )
    {
      t = KheTimeGroupTime(tg, j);
      count = KheResourceTimetableMonitorTimeTaskCount(rtm, t);
      for( k = 0;  k < count;  k++ )
      {
	task = KheResourceTimetableMonitorTimeTask(rtm, t, k);
	task = KheTaskProperRoot(task);
	if( task != NULL )
	  KheTaskUnVisit(task);
      }
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheFrameUnVisited(KHE_FRAME frame, KHE_RESOURCE r)                  */
/*                                                                           */
/*  Return true if all of r's tasks in frame are unvisited.  This will       */
/*  return true when there are no tasks in the frame.                        */
/*                                                                           */
/*****************************************************************************/

/* *** obsolete
bool KheFrameUnVisited(KHE_FRAME frame, KHE_RESOURCE r)
{
  KHE_RESOURCE_TIMETABLE_MONITOR rtm;  int i, j, k, count;
  KHE_POLARITY po;  KHE_TIME_GROUP tg;  KHE_TIME t;  KHE_TASK task;
  rtm = KheResourceTimetableMonitor(KheFrameSoln(frame), r);
  for( i = 0;  i < KheFrameTimeGroupCount(frame);  i++ )
  {
    tg = KheFrameTimeGroup(frame, i, &po);
    for( j = 0;  j < KheTimeGroupTimeCount(tg);  j++ )
    {
      t = KheTimeGroupTime(tg, j);
      count = KheResourceTimetableMonitorTimeTaskCount(rtm, t);
      for( k = 0;  k < count;  k++ )
      {
	task = KheResourceTimetableMonitorTimeTask(rtm, t, k);
	if( KheTaskVisited(task, 0) )
	  return false;
      }
    }
  }
  return true;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheFrameVisit(KHE_FRAME frame, KHE_RESOURCE r)                      */
/*                                                                           */
/*  Mark all of r's tasks in frame visited.                                  */
/*                                                                           */
/*****************************************************************************/

/* *** obsolete
void KheFrameVisit(KHE_FRAME frame, KHE_RESOURCE r)
{
  KHE_RESOURCE_TIMETABLE_MONITOR rtm;  int i, j, k, count;
  KHE_POLARITY po;  KHE_TIME_GROUP tg;  KHE_TIME t;  KHE_TASK task;
  rtm = KheResourceTimetableMonitor(KheFrameSoln(frame), r);
  for( i = 0;  i < KheFrameTimeGroupCount(frame);  i++ )
  {
    tg = KheFrameTimeGroup(frame, i, &po);
    for( j = 0;  j < KheTimeGroupTimeCount(tg);  j++ )
    {
      t = KheTimeGroupTime(tg, j);
      count = KheResourceTimetableMonitorTimeTaskCount(rtm, t);
      for( k = 0;  k < count;  k++ )
      {
	task = KheResourceTimetableMonitorTimeTask(rtm, t, k);
	task = KheTaskProperRoot(task)
	KheTaskVisit(task);
      }
    }
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheFrameUnVisit(KHE_FRAME frame, KHE_RESOURCE r)                    */
/*                                                                           */
/*  Mark all of r's tasks in frame unvisited.                                */
/*                                                                           */
/*****************************************************************************/

/* *** obsolete
void KheFrameUnVisit(KHE_FRAME frame, KHE_RESOURCE r)
{
  KHE_RESOURCE_TIMETABLE_MONITOR rtm;  int i, j, k, count;
  KHE_POLARITY po;  KHE_TIME_GROUP tg;  KHE_TIME t;  KHE_TASK task;
  rtm = KheResourceTimetableMonitor(KheFrameSoln(frame), r);
  for( i = 0;  i < KheFrameTimeGroupCount(frame);  i++ )
  {
    tg = KheFrameTimeGroup(frame, i, &po);
    for( j = 0;  j < KheTimeGroupTimeCount(tg);  j++ )
    {
      t = KheTimeGroupTime(tg, j);
      count = KheResourceTimetableMonitorTimeTaskCount(rtm, t);
      for( k = 0;  k < count;  k++ )
      {
	task = KheResourceTimetableMonitorTimeTask(rtm, t, k);
	KheTaskUnVisit(task);
      }
    }
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "frames - activity"                                            */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  bool KheTimeGroupIsActive(KHE_TIME_GROUP tg, KHE_POLARITY po,            */
/*    KHE_RESOURCE_TIMETABLE_MONITOR rtm)                                    */
/*                                                                           */
/*  Helper function which returns true when tg (whose polarity is po) is     */
/*  active for the resource monitored by rtm.                                */
/*                                                                           */
/*****************************************************************************/

static bool KheTimeGroupIsActive(KHE_TIME_GROUP tg, KHE_POLARITY po,
  KHE_RESOURCE_TIMETABLE_MONITOR rtm)
{
  return KheResourceTimetableMonitorTimeGroupAvailable(rtm, tg, false)
    != (po == KHE_POSITIVE);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheFrameIsActive(KHE_FRAME frame, KHE_RESOURCE r)                   */
/*                                                                           */
/*  Return true if frame is active for r.                                    */
/*                                                                           */
/*****************************************************************************/

bool KheFrameIsActive(KHE_FRAME frame, KHE_RESOURCE r)
{
  int i;  KHE_TIME_GROUP tg;  KHE_RESOURCE_TIMETABLE_MONITOR rtm;
  rtm = KheResourceTimetableMonitor(KheFrameSoln(frame), r);
  for( i = 0;  i < KheFrameTimeGroupCount(frame);  i++ )
  {
    tg = KheFrameTimeGroup(frame, i);
    if( !KheTimeGroupIsActive(tg, KHE_POSITIVE, rtm) )
      return false;
  }
  return true;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheFrameIsInactive(KHE_FRAME frame, KHE_RESOURCE r)                 */
/*                                                                           */
/*  Return true if frame is inactive for r.                                  */
/*                                                                           */
/*****************************************************************************/

bool KheFrameIsInactive(KHE_FRAME frame, KHE_RESOURCE r)
{
  int i;  KHE_TIME_GROUP tg;  KHE_RESOURCE_TIMETABLE_MONITOR rtm;
  rtm = KheResourceTimetableMonitor(KheFrameSoln(frame), r);
  for( i = 0;  i < KheFrameTimeGroupCount(frame);  i++ )
  {
    tg = KheFrameTimeGroup(frame, i);
    if( KheTimeGroupIsActive(tg, KHE_POSITIVE, rtm) )
      return false;
  }
  return true;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheFrameAdjacentToActive(KHE_FRAME frame, KHE_RESOURCE r)           */
/*                                                                           */
/*  Return true when frame is adjacent to an active index for r.             */
/*                                                                           */
/*****************************************************************************/

/* *** all correct but currently unused
bool KheFrameAdjacentToActive(KHE_FRAME frame, KHE_RESOURCE r)
{
  KHE_TIME_GROUP tg;  KHE_POLARITY po;  KHE_RESOURCE_TIMETABLE_MONITOR rtm;
  KHE_FRAME frame2;

  ** see whether there is an active index to the left **
  rtm = KheResourceTimetableMonitor(KheFrameSoln(frame), r);
  frame2 = KheFrameOutsideLeftSlice(frame, 1);
  if( KheFrameTimeGroupCount(frame2) > 0 )
  {
    tg = KheFrameTimeGroup(frame2, 0, &po);
    if( KheTimeGroupIsActive(tg, po, rtm) )
      return true;
  }

  ** see whether there is an active index to the right **
  frame2 = KheFrameOutsideRightSlice(frame, 1);
  if( KheFrameTimeGroupCount(frame2) > 0 )
  {
    tg = KheFrameTimeGroup(frame2, 0, &po);
    if( KheTimeGroupIsActive(tg, po, rtm) )
      return true;
  }

  ** no luck with either side **
  return false;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_FRAME KheFrameActiveAtLeft(KHE_FRAME frame, KHE_RESOURCE r)          */
/*                                                                           */
/*  Return the largest slice of frame which lies at the left end of frame    */
/*  and which is active for r.  If there is no such frame, the result will   */
/*  will be an empty frame.                                                  */
/*                                                                           */
/*****************************************************************************/

/* ***
KHE_FRAME KheFrameActiveAtLeft(KHE_FRAME frame, KHE_RESOURCE r)
{
  int i;  KHE_TIME_GROUP tg;  KHE_RESOURCE_TIMETABLE_MONITOR rtm;
  rtm = KheResourceTimetableMonitor(KheFrameSoln(frame), r);
  for( i = 0;  i < KheFrameTimeGroupCount(frame);  i++ )
  {
    tg = KheFrameTimeGroup(frame, i);
    if( !KheTimeGroupIsActive(tg, KHE_POSITIVE, rtm) )
      break;
  }
  return KheFrameInsideLeftSlice(frame, i);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_FRAME KheFrameInactiveAtLeft(KHE_FRAME frame, KHE_RESOURCE r)        */
/*                                                                           */
/*  Return the largest slice of frame which lies at the left end of frame    */
/*  and which is inactive for r.  If there is no such frame, the result will */
/*  be an empty frame.                                                       */
/*                                                                           */
/*****************************************************************************/

/* ***
KHE_FRAME KheFrameInactiveAtLeft(KHE_FRAME frame, KHE_RESOURCE r)
{
  int i;  KHE_TIME_GROUP tg;  KHE_RESOURCE_TIMETABLE_MONITOR rtm;
  rtm = KheResourceTimetableMonitor(KheFrameSoln(frame), r);
  for( i = 0;  i < KheFrameTimeGroupCount(frame);  i++ )
  {
    tg = KheFrameTimeGroup(frame, i);
    if( KheTimeGroupIsActive(tg, KHE_POSITIVE, rtm) )
      break;
  }
  return KheFrameInsideLeftSlice(frame, i);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_FRAME KheFrameActiveAtRight(KHE_FRAME frame, KHE_RESOURCE r)         */
/*                                                                           */
/*  Return the largest slice of frame which lies at the right end of frame   */
/*  and which is active for r.  If there is no such interval, the result     */
/*  will be an empty frame.                                                  */
/*                                                                           */
/*****************************************************************************/

/* ***
KHE_FRAME KheFrameActiveAtRight(KHE_FRAME frame, KHE_RESOURCE r)
{
  int i;  KHE_TIME_GROUP tg;  KHE_RESOURCE_TIMETABLE_MONITOR rtm;
  rtm = KheResourceTimetableMonitor(KheFrameSoln(frame), r);
  for( i = KheFrameTimeGroupCount(frame) - 1;  i >= 0;  i-- )
  {
    tg = KheFrameTimeGroup(frame, i);
    if( !KheTimeGroupIsActive(tg, KHE_POSITIVE, rtm) )
      break;
  }
  return KheFrameInsideRightSlice(frame, KheFrameTimeGroupCount(frame) - i - 1);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_FRAME KheFrameInactiveAtRight(KHE_FRAME frame, KHE_RESOURCE r)       */
/*                                                                           */
/*  Return the largest slice of frame which lies at the right end of frame   */
/*  and which is inactive for r.  If there is no such interval, the result   */
/*  will be an empty frame.                                                  */
/*                                                                           */
/*****************************************************************************/

/* ***
KHE_FRAME KheFrameInactiveAtRight(KHE_FRAME frame, KHE_RESOURCE r)
{
  int i;  KHE_TIME_GROUP tg;  KHE_RESOURCE_TIMETABLE_MONITOR rtm;
  rtm = KheResourceTimetableMonitor(KheFrameSoln(frame), r);
  for( i = KheFrameTimeGroupCount(frame) - 1;  i >= 0;  i-- )
  {
    tg = KheFrameTimeGroup(frame, i);
    if( KheTimeGroupIsActive(tg, KHE_POSITIVE, rtm) )
      break;
  }
  return KheFrameInsideRightSlice(frame, KheFrameTimeGroupCount(frame) - i - 1);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "frames - iteration"                                           */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  bool KheFrameIteratorCurrActive(KHE_FRAME_ITERATOR fi, int curr)         */
/*                                                                           */
/*  Return true if frame is active at curr.                                  */
/*                                                                           */
/*****************************************************************************/

/* ***
static bool KheFrameIteratorCurrActive(KHE_FRAME_ITERATOR fi, int curr)
{
  KHE_TIME_GROUP tg;  int index;
  index = (curr + fi->extra) % fi->count;
  if( index == fi->count - 1 )
    return false;
  tg = KheFrameTimeGroup(fi->frame, index);
  return KheTimeGroupIsActive(tg, KHE_POSITIVE, fi->rtm);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheFrameIteratorInit(KHE_FRAME_ITERATOR fi, KHE_FRAME frame,        */
/*    KHE_RESOURCE r, int extra)                                             */
/*                                                                           */
/*  Initialize frame iterator fi.                                            */
/*                                                                           */
/*****************************************************************************/

/* ***
void KheFrameIteratorInit(KHE_FRAME_ITERATOR fi, KHE_FRAME frame,
  KHE_RESOURCE r, int extra)
{
  ** do the basic initialization **
  fi->frame = frame;
  fi->count = KheFrameTimeGroupCount(frame) + 1;
  fi->extra = extra;
  fi->curr = 0;
  fi->resource = r;
  fi->rtm = KheResourceTimetableMonitor(KheFrameSoln(frame), r);

  ** ensure (fi->curr + fi->extra) % fi->count indexes an inactive time  **
  ** group for r.  There must be at least one inactive time group, since **
  ** count is 1 longer than the frame and the extra slot is inactive     **
  while( KheFrameIteratorCurrActive(fi, fi->curr) )
    fi->extra += 1;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheFrameIteratorNext(KHE_FRAME_ITERATOR fi, KHE_FRAME *res)         */
/*                                                                           */
/*  Return the next active frame found by frame iterator fi.                 */
/*                                                                           */
/*****************************************************************************/

/* ***
bool KheFrameIteratorNext(KHE_FRAME_ITERATOR fi, KHE_FRAME *res)
{
  int start_index, stop_index;

  ** skip over inactive elements.  It is a precondition that     **
  ** (fi->curr + fi->extra) % fi->count indexes an inactive time **
  ** group here at the start, so we don't need to check it       **
  do
  {
    fi->curr++;
    if( fi->curr == fi->count )
      return *res = KheFrameMakeNull(), false;
  } while( !KheFrameIteratorCurrActive(fi, fi->curr) );
  start_index = (fi->curr + fi->extra) % fi->count;

  ** skip over active elements **
  do
  {
    fi->curr++;
    if( fi->curr == fi->count )
      return *res = KheFrameMakeNull(), false;
  } while( KheFrameIteratorCurrActive(fi, fi->curr) );
  stop_index = (fi->curr + fi->extra) % fi->count;

  ** create the active frame and return true **
  *res = KheFrameSlice(fi->frame, start_index, stop_index);
  return true;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheFrameIteratorNext(KHE_FRAME_ITERATOR fi, KHE_FRAME *res)         */
/*                                                                           */
/*  Return the next pair of active frames (as a single frame) found by       */
/*  frame iterator fi.                                                       */
/*                                                                           */
/*****************************************************************************/

/* ***
bool KheFrameIteratorNextPair(KHE_FRAME_ITERATOR fi, KHE_FRAME *res)
{
  int start_index1, stop_index2, curr;

  do
  {
    ** skip over inactive elements.  It is a precondition that     **
    ** (fi->curr + fi->extra) % fi->count indexes an inactive time **
    ** group here at the start, so we don't need to check it       **
    do
    {
      fi->curr++;
      if( fi->curr == fi->count )
	return *res = KheFrameMakeNull(), false;
    } while( !KheFrameIteratorCurrActive(fi, fi->curr) );
    start_index1 = (fi->curr + fi->extra) % fi->count;

    ** skip over active elements **
    do
    {
      fi->curr++;
      if( fi->curr == fi->count )
	return *res = KheFrameMakeNull(), false;
    } while( KheFrameIteratorCurrActive(fi, fi->curr) );

    ** skip over inactive elements **
    curr = fi->curr;
    do
    {
      curr++;
      if( curr == fi->count )
	return *res = KheFrameMakeNull(), false;
    } while( !KheFrameIteratorCurrActive(fi, curr) );

    ** skip over active elements **
    do
    {
      curr++;
      if( curr == fi->count )
	return *res = KheFrameMakeNull(), false;
    } while( KheFrameIteratorCurrActive(fi, curr) );
    stop_index2 = (curr + fi->extra) % fi->count;

    ** if frames straddle the end, try again **
  } while( start_index1 >= stop_index2 );

  ** create the active frame and return true **
  *res = KheFrameSlice(fi->frame, start_index1, stop_index2);
  return true;
}
*** */
