
/*****************************************************************************/
/*                                                                           */
/*  THE KHE HIGH SCHOOL TIMETABLING ENGINE                                   */
/*  COPYRIGHT (C) 2010 Jeffrey H. Kingston                                   */
/*                                                                           */
/*  Jeffrey H. Kingston (jeff@it.usyd.edu.au)                                */
/*  School of Information Technologies                                       */
/*  The University of Sydney 2006                                            */
/*  AUSTRALIA                                                                */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU General Public License as published by     */
/*  the Free Software Foundation; either Version 3, or (at your option)      */
/*  any later version.                                                       */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston MA 02111-1307 USA   */
/*                                                                           */
/*  FILE:         khe_sr_group_by_frame.c                                    */
/*  DESCRIPTION:  Group by frame                                             */
/*                                                                           */
/*****************************************************************************/
#include "khe_solvers.h"
#include <limits.h>
#define bool_show(x) ((x) ? "true" : "false")

#define DEBUG1 0
#define DEBUG2 0
#define DEBUG3 1  /* shows which tasks are being grouped */
#define DEBUG4 0
#define DEBUG5 0
#define DEBUG6 0  /* phase 0 */
#define	DEBUG7 0  /* KhePlateauMerge */
#define	DEBUG8 0  /* KheMergeOneGap */
#define DEBUG9 0  /* KheGroupTwoMonitors */


/*****************************************************************************/
/*                                                                           */
/*  KHE_FRAMED_MONITOR - a monitor, plus a frame saying what it covers       */
/*                                                                           */
/*  Phase 0 applies when phase0_lim >= 0.                                    */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_framed_monitor_rec
{
  KHE_MONITOR		monitor;
  int			first_index;
  int			last_index;
  int			phase0_lim;
  HA_ARRAY_INT		phase0_intersects;
  int			phase0_supply;
  int			phase0_demand;
} *KHE_FRAMED_MONITOR;

typedef HA_ARRAY(KHE_FRAMED_MONITOR) ARRAY_KHE_FRAMED_MONITOR;


/*****************************************************************************/
/*                                                                           */
/*  KHE_TIME_SET - a set of times (like a time group, but without baggage)   */
/*                                                                           */
/*****************************************************************************/

/* *** moved to platform, khe_time_set.c
typedef HA_ARRAY(KHE_TIME) ARRAY_KHE_TIME;

typedef struct khe_time_set_rec
{
  ARRAY_KHE_TIME		times;
} *KHE_TIME_SET;
*** */

typedef HA_ARRAY(KHE_TIME_SET) ARRAY_KHE_TIME_SET;


/*****************************************************************************/
/*                                                                           */
/*  KHE_DOMAIN_TASK_SET - a set of tasks, broken by time, with one domain    */
/*                                                                           */
/*****************************************************************************/
typedef HA_ARRAY(KHE_TASK_SET) ARRAY_KHE_TASK_SET;

typedef struct khe_domain_task_set_rec {
  KHE_RESOURCE_GROUP			domain;		/* the shared domain */
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR	monitor;	/* the monitor       */
  int					min_limit;	/* min limit         */
  int					max_limit;	/* min limit         */
  ARRAY_KHE_TASK_SET			task_sets;	/* one per laim tg   */
} *KHE_DOMAIN_TASK_SET;

typedef HA_ARRAY(KHE_DOMAIN_TASK_SET) ARRAY_KHE_DOMAIN_TASK_SET;


/*****************************************************************************/
/*                                                                           */
/*  KHE_GROUP_BY_FRAME_SOLVER - solver for group by frame solving            */
/*                                                                           */
/*  NB start_indexes[si] is the smallest i such that                         */
/*  framed_monitors[i]->start_index >= si, or if there is no such            */
/*  index there is no entry for si in start_indexes.                         */
/*                                                                           */
/*  Consequently, the following code will visit all monitors whose           */
/*  frames start at or after si:                                             */
/*                                                                           */
/*  if( si < HaArrayCount(start_indexes) )                                   */
/*    for( i = start_indexes[si];  i < HaArrayCount(framed_monitors);  i++ ) */
/*      visit(framed_monitors[i];                                            */
/*                                                                           */
/*  and the first time the visit comes upon a monitor whose first index      */
/*  is past the interval we are interested in, we can stop.                  */
/*                                                                           */
/*****************************************************************************/

typedef HA_ARRAY(KHE_LIMIT_ACTIVE_INTERVALS_MONITOR)
  ARRAY_KHE_LIMIT_ACTIVE_INTERVALS_MONITOR;

typedef struct khe_group_by_frame_solver_rec {

  /* the solve as a whole */
  HA_ARENA			arena;
  KHE_TASKING			tasking;
  KHE_SOLN			soln;
  bool				resource_invariant;
  KHE_FRAME			common_frame;
  KHE_EVENT_TIMETABLE_MONITOR	etm;
  KHE_RESOURCE			resource;
  bool				soln_changed;
  KHE_TASK_SET			task_set;

  /* phase 0 */
  HA_ARRAY_BOOL			provisionally_linked_to_prev;
  HA_ARRAY_BOOL			linked_to_prev;

  /* phase 1 */
  int				max_days;
  ARRAY_KHE_FRAMED_MONITOR	framed_monitors;
  HA_ARRAY_INT			start_indexes;
  KHE_GROUP_MONITOR		group_monitor;
  KHE_TIME_SET			successful_times;
  int				last_index;
  int				success_count;
  KHE_TIME_SET			success_ts;
  KHE_TIME_SET			curr_ts;

  /* phase 2 */
  ARRAY_KHE_TIME_SET		time_sets;
  ARRAY_KHE_TIME_SET		time_sets_by_time;
  ARRAY_KHE_TIME_SET		time_sets_by_monitor;
  ARRAY_KHE_LIMIT_ACTIVE_INTERVALS_MONITOR suitable_monitors;
  HA_ARRAY_INT			success_seq;
  HA_ARRAY_INT			curr_seq;
  int				node_count;
  ARRAY_KHE_TASK_SET		task_sets;
  HA_ARRAY_INT			counts;
  ARRAY_KHE_DOMAIN_TASK_SET	domain_task_sets;

} *KHE_GROUP_BY_FRAME_SOLVER;


/*****************************************************************************/
/*                                                                           */
/*  Submodule "domain task sets"                                             */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DOMAIN_TASK_SET KheDomainTaskSetMake(KHE_GROUP_BY_FRAME_SOLVER gfs,  */
/*    KHE_LIMIT_ACTIVE_INTERVALS_MONITOR monitor, KHE_RESOURCE_GROUP domain, */
/*    int min_limit, int max_limit)                                          */
/*                                                                           */
/*  Make a new domain task set with these attributes.                        */
/*                                                                           */
/*****************************************************************************/

static KHE_DOMAIN_TASK_SET KheDomainTaskSetMake(KHE_GROUP_BY_FRAME_SOLVER gfs,
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR monitor, KHE_RESOURCE_GROUP domain,
  int min_limit, int max_limit)
{
  KHE_DOMAIN_TASK_SET res;  int i, tg_count;  KHE_TASK_SET ts;
  HaMake(res, gfs->arena);
  res->domain = domain;
  res->monitor = monitor;
  res->min_limit = min_limit;
  res->max_limit = max_limit;
  HaArrayInit(res->task_sets, gfs->arena);
  tg_count = KheLimitActiveIntervalsMonitorTimeGroupCount(monitor);
  for( i = 0;  i < tg_count;  i++ )
  {
    ts = KheTaskSetMake(gfs->soln);
    HaArrayAddLast(res->task_sets, ts);
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDomainTaskSetAddTask(KHE_DOMAIN_TASK_SET dts, KHE_TASK task,     */
/*    int index)                                                             */
/*                                                                           */
/*  Add task to dts, at the appropriate index.                               */
/*                                                                           */
/*****************************************************************************/

static void KheDomainTaskSetAddTask(KHE_DOMAIN_TASK_SET dts, KHE_TASK task,
  int index)
{
  KHE_TASK_SET ts;
  HnAssert(0 <= index && index < HaArrayCount(dts->task_sets),
    "KheDomainTaskSetAddTask: index (%d) out of range (0 .. %d)",
    index, HaArrayCount(dts->task_sets) - 1);
  ts = HaArray(dts->task_sets, index);
  HnAssert(ts != NULL, "KheDomainTaskSetAddTask internal error");
  KheTaskSetAddTask(ts, task);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDomainTaskSetFindPlateau(KHE_DOMAIN_TASK_SET dts,                */
/*    int *start_pos, int *end_pos)                                          */
/*                                                                           */
/*  If dts contains a plateau, return the first one.                         */
/*                                                                           */
/*  Writing Num(i) for the number of tasks at position i, a plateau          */
/*  is a pair of indexes (start_pos, end_pos) satisfying                     */
/*                                                                           */
/*    Num(start_pos - 1) < Num(start_pos) &&                                 */
/*    Num(start_pos) == Num(start_pos + 1) == ... == Num(end_pos) &&         */
/*    Num(end_pos) > Num(end_pos + 1)                                        */
/*                                                                           */
/*  This means it is possible to group tasks at start_pos ... end_pos,       */
/*  although whether grouping will actually be possible depends on the       */
/*  length of the plateau (end_pos - start_pos + 1) as well.                 */
/*                                                                           */
/*****************************************************************************/

static bool KheDomainTaskSetFindPlateau(KHE_DOMAIN_TASK_SET dts,
  int *start_pos, int *end_pos)
{
  KHE_TASK_SET ts1, ts2;  int i, prev_up;
  prev_up = -1;
  for( i = 0;  i < HaArrayCount(dts->task_sets) - 1;  i++ )
  {
    ts1 = HaArray(dts->task_sets, i);
    ts2 = HaArray(dts->task_sets, i + 1);
    if( KheTaskSetTaskCount(ts1) < KheTaskSetTaskCount(ts2) )
      prev_up = i + 1;
    else if( KheTaskSetTaskCount(ts1) > KheTaskSetTaskCount(ts2) )
    {
      if( prev_up != -1 )
	return *start_pos = prev_up, *end_pos = i, true;
    }
  }
  return *start_pos = -1, *end_pos = -1, false;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDomainTaskSetFindTaskRange(KHE_DOMAIN_TASK_SET dts,              */
/*    KHE_TASK task, int start_pos, int *end_pos)                            */
/*                                                                           */
/*  Assuming that task is present at start_pos, set *end_pos to the end      */
/*  point of its range.                                                      */
/*                                                                           */
/*****************************************************************************/

static void KheDomainTaskSetFindTaskRange(KHE_DOMAIN_TASK_SET dts,
  KHE_TASK task, int start_pos, int *end_pos)
{
  int i;
  for( i = start_pos + 1;  i <= 
  {
    ts = HaArray(dts->task_sets, j);
    if( !KheTaskSetContainsTask(ts, task, &pos) )
      break;
  }

}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDomainTaskSetContainsGroupableTask(KHE_DOMAIN_TASK_SET dts,      */
/*    int *start_pos, int end_pos, KHE_TASK *res)                            */
/*                                                                           */
/*  Search for a task that covers *start_pos ... end_pos, update them        */
/*  and return the task.                                                     */
/*                                                                           */
/*****************************************************************************/

bool KheDomainTaskSetContainsGroupableTask(KHE_DOMAIN_TASK_SET dts,
  int *start_pos, int end_pos, KHE_TASK *res)
{
  KHE_TASK_SET ts;  int i, j, k, pos;  KHE_TASK task;
  ts = HaArray(dts->task_sets, *start_pos);
  for( i = 0;  i < KheTaskSetTaskCount(ts);  i++ )
  {
    /* see whether task is groupable */
    task = KheTaskSetTask(ts, i);

    /* set *start_pos ... j - 1 to the range covered by task */
    for( j = *start_pos + 1;  j <= *end_pos;  j++ )
    {
      ts = HaArray(dts->task_sets, j);
      if( !KheTaskSetContainsTask(ts, task, &pos) )
	break;
    }

    /* if this range equals the total duration of task, return it */
    if( j - *start_pos == KheTaskTotalDuration(task) )
    {
      for( k = *start_pos;  k < j;  k++ )
      {
	ts = HaArray(dts->task_sets, k);
	KheTaskSetDeleteTask(ts, task);
      }
      return *start_pos = j, *res = task, true;
    }
  }
  return *start_pos = -1, *res = NULL, false;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDomainTaskSetGroup(KHE_DOMAIN_TASK_SET dts,                      */
/*    int start_pos, int end_pos, KHE_GROUP_BY_FRAME_SOLVER gfs)             */
/*                                                                           */
/*  Remove one task from each task set in this range and group them.  Be     */
/*  careful to take account of existing groupings:  if a task is part of     */
/*  a group that extends outside this range, don't use it; and if it         */
/*  covers several indexes within the range, treat those as all done.        */
/*                                                                           */
/*  Any tasks that don't work get removed anyway.                            */
/*                                                                           */
/*****************************************************************************/

static void KheDomainTaskSetGroup(KHE_DOMAIN_TASK_SET dts,
  int start_pos, int end_pos, KHE_GROUP_BY_FRAME_SOLVER gfs)
{
  KHE_TASK_SET ts;  int i;  KHE_TASK first_task, task;  bool ok;
  first_task = NULL;  starting = true;
  while( KheDomainTaskSetContainsGroupableTask(dts, &start_pos, end_pos, &task))
  {
    if( first_task == NULL )
      first_task = task;
    else
    {
      ok = KheTaskAssign(task, first_task);
      if( ok )
      {
	if( DEBUG3 )
	{
	  if( starting )
	  {
	    fprintf(stderr, "    KheGroupByFrame plateau grouping tasks {");
	    KheTaskDebug(first_task, 2, -1, stderr);
	    starting = false;
	  }
	  fprintf(stderr, ", ");
	  KheTaskDebug(task, 2, -1, stderr);
	}
	gfs->soln_changed = true;
	if( gfs->task_set != NULL )
	  KheTaskSetAddTask(gfs->task_set, task);
      }
    }
  }
  if( DEBUG3 && !starting )
    fprintf(stderr, "}\n");
}


/* *** old version that does not take existing groupings into account
static void KheDomainTaskSetGroup(KHE_DOMAIN_TASK_SET dts,
  int start_pos, int end_pos, KHE_GROUP_BY_FRAME_SOLVER gfs)
{
  KHE_TASK_SET ts;  int i;  KHE_TASK first_task, task;  bool ok;
  ts = HaArray(dts->task_sets, start_pos);
  first_task = KheTaskSetTask(ts, KheTaskSetTaskCount(ts) - 1);
  if( DEBUG3 )
  {
    fprintf(stderr, "    KheGroupByFrame plateau grouping tasks {");
    KheTaskDebug(first_task, 1, -1, stderr);
  }
  for( i = start_pos + 1;  i <= end_pos;  i++ )
  {
    ts = HaArray(dts->task_sets, i);
    task = KheTaskSetTask(ts, KheTaskSetTaskCount(ts) - 1);
    if( task == first_task )
      ok = true;
    else
    {
      ok = KheTaskAssign(task, first_task);
      if( ok )
      {
	gfs->soln_changed = true;
	if( gfs->task_set != NULL )
	  KheTaskSetAddTask(gfs->task_set, task);
      }
    }
    if( DEBUG3 )
    {
      fprintf(stderr, ", %s", ok ? "" : "not ");
      KheTaskDebug(task, 1, -1, stderr);
    }
  }
  if( DEBUG3 )
    fprintf(stderr, "}\n");
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDomainTaskSetGroupAndRemovePlateau(KHE_DOMAIN_TASK_SET dts,      */
/*    int start_pos, int end_pos, KHE_GROUP_BY_FRAME_SOLVER gfs)             */
/*                                                                           */
/*  There is a plateau at start_pos .. end_pos.  Group it and remove it.     */
/*                                                                           */
/*****************************************************************************/

static void KheDomainTaskSetGroupAndRemovePlateau(KHE_DOMAIN_TASK_SET dts,
  int start_pos, int end_pos, KHE_GROUP_BY_FRAME_SOLVER gfs)
{
  int i, len;  KHE_TASK_SET ts;

  /* optionally group the plateau */
  len = end_pos - start_pos + 1;
  if( len == 1 || len > dts->max_limit )
  {
    /* too small or oo large to group */
  }
  else if( len < 2 * dts->min_limit )
  {
    /* can't break in two, so group it all */
    KheDomainTaskSetGroup(dts, start_pos, end_pos, gfs);
  }
  else if( len == 2 * dts->min_limit )
  {
    /* two equal groups is the only option here */
    /* given that they can both be assigned to the same resource */
    KheDomainTaskSetGroup(dts, start_pos, start_pos + dts->min_limit - 1, gfs);
    KheDomainTaskSetGroup(dts, start_pos + dts->min_limit, end_pos, gfs);
  }
  else
  {
    /* probably no single solution, so leave it ungrouped */
  }

  /* remove the plateau */
  for( i = start_pos;  i <= end_pos;  i++ )
  {
    ts = HaArray(dts->task_sets, i);
    HnAssert(KheTaskSetTaskCount(ts) > 0,
      "KheDomainTaskSetGroupAndRemovePlateau internal error");
    KheTaskSetDropFromEnd(ts, 1);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDomainTaskSetPlateauDebug(KHE_DOMAIN_TASK_SET dts,               */
/*    int start_pos, int end_pos, int verbosity, int indent, FILE *fp)       */
/*                                                                           */
/*  Debug print of a plateau of dts at the given verbosity.                  */
/*                                                                           */
/*****************************************************************************/

static void KheDomainTaskSetPlateauDebug(KHE_DOMAIN_TASK_SET dts,
  int start_pos, int end_pos, int verbosity, int indent, FILE *fp)
{
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT laic; KHE_TIME t1, t2;
  KHE_TIME_GROUP tg;  KHE_POLARITY po;
  laic = KheLimitActiveIntervalsMonitorConstraint(dts->monitor);
  tg = KheLimitActiveIntervalsMonitorTimeGroup(dts->monitor, start_pos, &po);
  t1 = KheTimeGroupTime(tg, 0);
  tg = KheLimitActiveIntervalsMonitorTimeGroup(dts->monitor, end_pos, &po);
  t2 = KheTimeGroupTime(tg, 0);
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  fprintf(fp, "Plateau(%s at %s .. %s)",
    KheConstraintId((KHE_CONSTRAINT) laic), KheTimeId(t1), KheTimeId(t2));
  if( indent > 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDomainTaskSetDebug(KHE_DOMAIN_TASK_SET dts, int verbosity,       */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Debug print of dts onto fp with the given verbosity and indent.          */
/*                                                                           */
/*****************************************************************************/

static void KheDomainTaskSetDebug(KHE_DOMAIN_TASK_SET dts, int verbosity,
  int indent, FILE *fp)
{
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT laic; KHE_TIME t; KHE_TASK_SET ts;
  KHE_TIME_GROUP tg;  KHE_POLARITY po;  int i, tg_count;
  fprintf(fp, "%*s[ DomainTaskSet(", indent, "");
  laic = KheLimitActiveIntervalsMonitorConstraint(dts->monitor);
  fprintf(fp, "%s, %d-%d)\n", KheConstraintId((KHE_CONSTRAINT) laic),
    dts->min_limit, dts->max_limit);
  KheResourceGroupDebug(dts->domain, verbosity, indent + 2, fp);
  tg_count = KheLimitActiveIntervalsMonitorTimeGroupCount(dts->monitor);
  for( i = 0;  i < tg_count;  i++ )
  {
    tg = KheLimitActiveIntervalsMonitorTimeGroup(dts->monitor, i, &po);
    HnAssert(KheTimeGroupTimeCount(tg) == 1 && po == KHE_POSITIVE,
      "KheDomainTaskSetDebug internal error 1");
    t = KheTimeGroupTime(tg, 0);
    ts = HaArray(dts->task_sets, i);
    HnAssert(ts != NULL, "KheDomainTaskSetDebug internal error 2");
    fprintf(fp, "%*s  %s (%d): ", indent, "", KheTimeId(t),
      KheTaskSetTaskCount(ts));
    KheTaskSetDebug(ts, verbosity, 0, fp);
  }
  fprintf(fp, "%*s]\n", indent, "");
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheContainsDomainTaskSet(KHE_GROUP_BY_FRAME_SOLVER gfs,             */
/*    KHE_RESOURCE_GROUP rg, KHE_DOMAIN_TASK_SET *res)                       */
/*                                                                           */
/*  If gfs contains a domain task set with domain rg, set *res to that       */
/*  domain task set and return true.  Otherwise return false.                */
/*                                                                           */
/*****************************************************************************/

static bool KheContainsDomainTaskSet(KHE_GROUP_BY_FRAME_SOLVER gfs,
  KHE_RESOURCE_GROUP rg, KHE_DOMAIN_TASK_SET *res)
{
  KHE_DOMAIN_TASK_SET dts;  int i;
  HaArrayForEach(gfs->domain_task_sets, dts, i)
    if( KheResourceGroupEqual(dts->domain, rg) )
      return *res = dts, true;
  return *res = NULL, false;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheMakeDomainTaskSets(KHE_GROUP_BY_FRAME_SOLVER gfs,                */
/*    KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim, int min_limit, int max_limit) */
/*                                                                           */
/*  Make domain task sets for laim.                                          */
/*                                                                           */
/*****************************************************************************/

static void KheMakeDomainTaskSets(KHE_GROUP_BY_FRAME_SOLVER gfs,
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim, int min_limit, int max_limit)
{
  KHE_TIME t;  KHE_TIME_GROUP tg;  int i, j, k, tg_count;  KHE_POLARITY po;
  KHE_MEET meet;  KHE_TASK task;  KHE_DOMAIN_TASK_SET dts;
  KHE_RESOURCE_TYPE rt;

  /* boilerplate */
  HaArrayClear(gfs->domain_task_sets);
  rt = KheResourceResourceType(gfs->resource);

  /* find and add each task of the appropriate type running under laim */
  tg_count = KheLimitActiveIntervalsMonitorTimeGroupCount(laim);
  for( i = 0;  i < tg_count;  i++ )
  {
    tg = KheLimitActiveIntervalsMonitorTimeGroup(laim, i, &po);
    HnAssert(KheTimeGroupTimeCount(tg) == 1 && po == KHE_POSITIVE,
      "KheMakeDomainTaskSets internal error 1");
    t = KheTimeGroupTime(tg, 0);
    for( j = 0;  j < KheEventTimetableMonitorTimeMeetCount(gfs->etm, t);  j++ )
    {
      meet = KheEventTimetableMonitorTimeMeet(gfs->etm, t, j);
      if( KheMeetDuration(meet) == 1 )
	for( k = 0;  k < KheMeetTaskCount(meet);  k++ )
	{
	  task = KheProperRoot(KheMeetTask(meet, k));
	  if( KheTaskResourceType(task) == rt &&
	      KheTaskNonAssignmentHasCost(task) )
	  {
	    /* add task to the appropriate domain task set */
            if( !KheContainsDomainTaskSet(gfs, KheTaskDomain(task), &dts) )
	    {
	      dts = KheDomainTaskSetMake(gfs, laim, KheTaskDomain(task),
		min_limit, max_limit);
	      HaArrayAddLast(gfs->domain_task_sets, dts);
	    }
	    KheDomainTaskSetAddTask(dts, task, i);
	  }
	}
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "framed monitors"                                              */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_FRAMED_MONITOR KheFramedMonitorMake(KHE_MONITOR m,                   */
/*    int first_index, int last_index)                                       */
/*                                                                           */
/*  Make a new framed monitor object with these attributes.                  */
/*                                                                           */
/*****************************************************************************/

static KHE_FRAMED_MONITOR KheFramedMonitorMake(KHE_MONITOR m,
  int first_index, int last_index, HA_ARENA a)
{
  KHE_FRAMED_MONITOR res;
  HaMake(res, a);
  res->monitor = m;
  res->first_index = first_index;
  res->last_index = last_index;
  res->phase0_lim = -1;
  HaArrayInit(res->phase0_intersects, a);
  res->phase0_supply = res->phase0_demand = -1;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheFramedMonitorCmp(const void *t1, const void *t2)                  */
/*                                                                           */
/*  Untyped comparison function for sorting an array of framed monitor       */
/*  objects by increasing (first_index, last_index).                         */
/*                                                                           */
/*****************************************************************************/

int KheFramedMonitorCmp(const void *t1, const void *t2)
{
  KHE_FRAMED_MONITOR fm1 = * (KHE_FRAMED_MONITOR *) t1;
  KHE_FRAMED_MONITOR fm2 = * (KHE_FRAMED_MONITOR *) t2;
  int cmp;
  cmp = fm1->first_index - fm2->first_index;
  if( cmp != 0 ) return cmp;
  return fm1->last_index - fm2->last_index;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheClusterBusyTimesConstraintSingleResourceType(                    */
/*    KHE_CLUSTER_BUSY_TIMES_CONSTRAINT c, KHE_RESOURCE_TYPE *rt)            */
/*                                                                           */
/*  If the resources that c applies to have a single resource type, set      */
/*  *rt to that resource type and return true.  Otherwise return false.      */
/*                                                                           */
/*****************************************************************************/

static bool KheClusterBusyTimesConstraintSingleResourceType(
  KHE_CLUSTER_BUSY_TIMES_CONSTRAINT c, KHE_RESOURCE_TYPE *rt)
{
  int i;  KHE_RESOURCE_GROUP rg;  KHE_RESOURCE r;
  *rt = NULL;
  for( i = 0;  i < KheClusterBusyTimesConstraintResourceGroupCount(c);  i++ )
  {
    rg = KheClusterBusyTimesConstraintResourceGroup(c, i);
    if( *rt == NULL )
      *rt = KheResourceGroupResourceType(rg);
    else if( *rt != KheResourceGroupResourceType(rg) )
      return *rt = NULL, false;
  }
  for( i = 0;  i < KheClusterBusyTimesConstraintResourceCount(c);  i++ )
  {
    r = KheClusterBusyTimesConstraintResource(c, i);
    if( *rt == NULL )
      *rt = KheResourceResourceType(r);
    else if( *rt != KheResourceResourceType(r) )
      return *rt = NULL, false;
  }
  return *rt != NULL;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheClusterBusyTimesConstraintCount(                                  */
/*    KHE_CLUSTER_BUSY_TIMES_CONSTRAINT c)                                   */
/*                                                                           */
/*  Return the number of resources monitored by c.                           */
/*                                                                           */
/*****************************************************************************/

static int KheClusterBusyTimesConstraintCount(
  KHE_CLUSTER_BUSY_TIMES_CONSTRAINT c)
{
  int i, count;  KHE_RESOURCE_GROUP rg;
  count = KheClusterBusyTimesConstraintResourceCount(c);
  for( i = 0;  i < KheClusterBusyTimesConstraintResourceGroupCount(c);  i++ )
  {
    rg = KheClusterBusyTimesConstraintResourceGroup(c, i);
    count += KheResourceGroupResourceCount(rg);
  }
  return count;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheUseClusterBusyTimesMonitor(KHE_CLUSTER_BUSY_TIMES_MONITOR cbtm,  */
/*    KHE_FRAME common_frame, KHE_FRAMED_MONITOR *res)                       */
/*                                                                           */
/*  If cbtm is relevant, return true and set *res to its framed monitor      */
/*  object.  Otherwise return false.                                         */
/*                                                                           */
/*****************************************************************************/

static bool KheUseClusterBusyTimesMonitor(KHE_CLUSTER_BUSY_TIMES_MONITOR cbtm,
  KHE_FRAME common_frame, KHE_FRAMED_MONITOR *res, HA_ARENA a)
{
  int i, c_count;  KHE_TIME_GROUP tg;  KHE_POLARITY po;  KHE_RESOURCE_TYPE rt;
  KHE_TIME t, earliest_t, latest_t;  KHE_CLUSTER_BUSY_TIMES_CONSTRAINT cbtc;

  /* if monitor's constraint does not monitor most resources, return false */
  cbtc = KheClusterBusyTimesMonitorConstraint(cbtm);
  if( !KheClusterBusyTimesConstraintSingleResourceType(cbtc, &rt) )
    return *res = NULL, false;
  c_count = KheClusterBusyTimesConstraintCount(cbtc);
  if( c_count < 0.8 * KheResourceTypeResourceCount(rt) )
    return *res = NULL, false;
  /* ***
  if( KheClusterBusyTimesConstraintResourceGroupCount(cbtc) != 1 )
    return *res = NULL, false;
  if( KheResourceGroupResourceCount(rg) !=
      KheResourceTypeResourceCount(KheResourceGroupResourceType(rg)) )
    return *res = NULL, false;
  *** */

  /* find the earliest and latest times; if none, return false */
  earliest_t = latest_t = NULL;
  for( i = 0;  i < KheClusterBusyTimesMonitorTimeGroupCount(cbtm);  i++ )
  {
    tg = KheClusterBusyTimesMonitorTimeGroup(cbtm, i, &po);
    if( KheTimeGroupTimeCount(tg) > 0 )
    {
      t = KheTimeGroupTime(tg, 0);
      if( earliest_t == NULL || KheTimeIndex(t) < KheTimeIndex(earliest_t) )
        earliest_t = t;
      t = KheTimeGroupTime(tg, KheTimeGroupTimeCount(tg) - 1);
      if( latest_t == NULL || KheTimeIndex(t) > KheTimeIndex(latest_t) )
        latest_t = t;
    }
  }
  if( earliest_t == NULL )
    return *res = NULL, false;

  /* all good, make and return a framed monitor object */
  *res = KheFramedMonitorMake((KHE_MONITOR) cbtm,
    KheFrameTimeIndex(common_frame, earliest_t),
    KheFrameTimeIndex(common_frame, latest_t), a);
    
  return true;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheLimitBusyTimesConstraintSingleResourceType(                      */
/*    KHE_LIMIT_BUSY_TIMES_CONSTRAINT c, KHE_RESOURCE_TYPE *rt)              */
/*                                                                           */
/*  If the resources that c applies to have a single resource type, set      */
/*  *rt to that resource type and return true.  Otherwise return false.      */
/*                                                                           */
/*****************************************************************************/

static bool KheLimitBusyTimesConstraintSingleResourceType(
  KHE_LIMIT_BUSY_TIMES_CONSTRAINT c, KHE_RESOURCE_TYPE *rt)
{
  int i;  KHE_RESOURCE_GROUP rg;  KHE_RESOURCE r;
  *rt = NULL;
  for( i = 0;  i < KheLimitBusyTimesConstraintResourceGroupCount(c);  i++ )
  {
    rg = KheLimitBusyTimesConstraintResourceGroup(c, i);
    if( *rt == NULL )
      *rt = KheResourceGroupResourceType(rg);
    else if( *rt != KheResourceGroupResourceType(rg) )
      return *rt = NULL, false;
  }
  for( i = 0;  i < KheLimitBusyTimesConstraintResourceCount(c);  i++ )
  {
    r = KheLimitBusyTimesConstraintResource(c, i);
    if( *rt == NULL )
      *rt = KheResourceResourceType(r);
    else if( *rt != KheResourceResourceType(r) )
      return *rt = NULL, false;
  }
  return *rt != NULL;
}

/*****************************************************************************/
/*                                                                           */
/*  int KheLimitBusyTimesConstraintCount(                                    */
/*    KHE_LIMIT_BUSY_TIMES_CONSTRAINT c)                                     */
/*                                                                           */
/*  Return the number of resources monitored by c.                           */
/*                                                                           */
/*****************************************************************************/

static int KheLimitBusyTimesConstraintCount(
  KHE_LIMIT_BUSY_TIMES_CONSTRAINT c)
{
  int i, count;  KHE_RESOURCE_GROUP rg;
  count = KheLimitBusyTimesConstraintResourceCount(c);
  for( i = 0;  i < KheLimitBusyTimesConstraintResourceGroupCount(c);  i++ )
  {
    rg = KheLimitBusyTimesConstraintResourceGroup(c, i);
    count += KheResourceGroupResourceCount(rg);
  }
  return count;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheUseLimitBusyTimesMonitor(KHE_LIMIT_BUSY_TIMES_MONITOR lbtm,      */
/*    KHE_FRAME common_frame, KHE_FRAMED_MONITOR *res)                       */
/*                                                                           */
/*  If lbtm is relevant, return true and set *res to its framed monitor      */
/*  object.  Otherwise return false.                                         */
/*                                                                           */
/*****************************************************************************/

static bool KheUseLimitBusyTimesMonitor(KHE_LIMIT_BUSY_TIMES_MONITOR lbtm,
  KHE_FRAME common_frame, KHE_FRAMED_MONITOR *res, HA_ARENA a)
{
  int i, offset, c_count;  KHE_TIME_GROUP tg;  KHE_RESOURCE_TYPE rt;
  KHE_TIME t, earliest_t, latest_t;  KHE_LIMIT_BUSY_TIMES_CONSTRAINT lbtc;

  /* if monitor's constraint does not monitor all resources, return false */
  lbtc = KheLimitBusyTimesMonitorConstraint(lbtm);
  if( !KheLimitBusyTimesConstraintSingleResourceType(lbtc, &rt) )
    return *res = NULL, false;
  c_count = KheLimitBusyTimesConstraintCount(lbtc);
  if( c_count < 0.8 * KheResourceTypeResourceCount(rt) )
    return *res = NULL, false;
  /* ***
  if( KheLimitBusyTimesConstraintResourceGroupCount(lbtc) != 1 )
    return *res = NULL, false;
  rg = KheLimitBusyTimesConstraintResourceGroup(lbtc, 0);
  if( KheResourceGroupResourceCount(rg) !=
      KheResourceTypeResourceCount(KheResourceGroupResourceType(rg)) )
    return *res = NULL, false;
  *** */

  /* find the earliest and latest times; if none, return false */
  earliest_t = latest_t = NULL;
  offset = KheLimitBusyTimesMonitorOffset(lbtm);
  for( i = 0;  i < KheLimitBusyTimesConstraintTimeGroupCount(lbtc);  i++ )
  {
    tg = KheLimitBusyTimesConstraintTimeGroup(lbtc, i, offset);
    if( KheTimeGroupTimeCount(tg) > 0 )
    {
      t = KheTimeGroupTime(tg, 0);
      if( earliest_t == NULL || KheTimeIndex(t) < KheTimeIndex(earliest_t) )
        earliest_t = t;
      t = KheTimeGroupTime(tg, KheTimeGroupTimeCount(tg) - 1);
      if( latest_t == NULL || KheTimeIndex(t) > KheTimeIndex(latest_t) )
        latest_t = t;
    }
  }
  if( earliest_t == NULL )
    return *res = NULL, false;

  /* all good, return a framed monitor object */
  return *res = KheFramedMonitorMake((KHE_MONITOR) lbtm,
    KheFrameTimeIndex(common_frame, earliest_t),
    KheFrameTimeIndex(common_frame, latest_t), a), true;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheUseMonitor(KHE_MONITOR m, KHE_FRAME common_frame,                */
/*    KHE_FRAMED_MONITOR *res)                                               */
/*                                                                           */
/*  If m is relevant, return true and set *res to its framed monitor         */
/*  object.  Otherwise return false.                                         */
/*                                                                           */
/*****************************************************************************/

static bool KheUseMonitor(KHE_MONITOR m, KHE_FRAME common_frame,
  KHE_FRAMED_MONITOR *res, HA_ARENA a)
{
  switch( KheMonitorTag(m) )
  {
    case KHE_CLUSTER_BUSY_TIMES_MONITOR_TAG:

      return KheUseClusterBusyTimesMonitor((KHE_CLUSTER_BUSY_TIMES_MONITOR) m,
        common_frame, res, a);

    case KHE_LIMIT_BUSY_TIMES_MONITOR_TAG:

      return KheUseLimitBusyTimesMonitor((KHE_LIMIT_BUSY_TIMES_MONITOR) m,
        common_frame, res, a);

    default:

      /* ignore other kinds of monitors */
      return *res = NULL, false;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheFramedMonitorDebug(KHE_FRAMED_MONITOR fm, int indent, FILE *fp)  */
/*                                                                           */
/*  Debug print of fm onto fp with the given indent.                         */
/*                                                                           */
/*****************************************************************************/

static void KheFramedMonitorDebug(KHE_FRAMED_MONITOR fm, int indent, FILE *fp)
{
  int val, i;  KHE_CLUSTER_BUSY_TIMES_MONITOR cbtm;  KHE_POLARITY po;
  KHE_TIME_GROUP tg;
  if( indent >= 0 )
  {
    fprintf(fp, "%*s[ Framed monitor %d-%d\n", indent, "",
      fm->first_index, fm->last_index);
    KheMonitorDebug(fm->monitor, 2, indent + 2, fp);
    if( fm->phase0_lim >= 0 )
    {
      cbtm = (KHE_CLUSTER_BUSY_TIMES_MONITOR) fm->monitor;
      HaArrayForEach(fm->phase0_intersects, val, i)
      {
	fprintf(fp, "%*s  %d time groups of frame intersect with ",
	  indent, "", val);
	tg = KheClusterBusyTimesMonitorTimeGroup(cbtm, i, &po);
	KheTimeGroupDebug(tg, 1, 0, fp);
      }
      fprintf(fp, "%*s  supply %d, demand %d\n", indent, "",
        fm->phase0_supply, fm->phase0_demand);
    }
    fprintf(fp, "%*s]\n", indent, "");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "time sets"                                                    */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_TIME_SET KheTimeSetMake(void)                                        */
/*                                                                           */
/*  Make a new, empty time set object.                                       */
/*                                                                           */
/*****************************************************************************/

/* *** moved to platform, khe_time_set.c
static KHE_TIME_SET KheTimeSetMake(HA_ARENA a)
{
  KHE_TIME_SET res;
  HaMake(res, a);
  HaArrayInit(res->times, a);
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  int KheTimeSetTimeCount(KHE_TIME_SET ts)                                 */
/*                                                                           */
/*  Return the number of times in ts.                                        */
/*                                                                           */
/*****************************************************************************/

/* *** moved to platform, khe_time_set.c
static int KheTimeSetTimeCount(KHE_TIME_SET ts)
{
  return HaArrayCount(ts->times);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheTimeSetClear(KHE_TIME_SET ts)                                    */
/*                                                                           */
/*  Clear (but do not delete) ts.                                            */
/*                                                                           */
/*****************************************************************************/

/* *** moved to platform, khe_time_set.c
static void KheTimeSetClear(KHE_TIME_SET ts)
{
  HaArrayClear(ts->times);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheTimeSetAddTime(KHE_TIME_SET ts, KHE_TIME t)                      */
/*                                                                           */
/*  Add t to ts.                                                             */
/*                                                                           */
/*****************************************************************************/

/* *** moved to platform, khe_time_set.c
static void KheTimeSetAddTime(KHE_TIME_SET ts, KHE_TIME t)
{
  HaArrayAddLast(ts->times, t);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheTimeSetContainsTime(KHE_TIME_SET ts, KHE_TIME t)                 */
/*                                                                           */
/*  Return true if ts contains t.                                            */
/*                                                                           */
/*****************************************************************************/

/* *** moved to platform, khe_time_set.c
static bool KheTimeSetContainsTime(KHE_TIME_SET ts, KHE_TIME t)
{
  int pos;
  return HaArrayContains(ts->times, t, &pos);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheTimeSetCopy(KHE_TIME_SET target, KHE_TIME_SET source)            */
/*                                                                           */
/*  Clear out whatever was in target, and copy source into it.               */
/*                                                                           */
/*****************************************************************************/

/* *** moved to platform, khe_time_set.c
static void KheTimeSetCopy(KHE_TIME_SET target, KHE_TIME_SET source)
{
  int i;
  HaArrayClear(target->times);
  HaArrayAppend(target->times, source->times, i);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheTimeSetDebug(KHE_TIME_SET ts, int indent, FILE *fp)              */
/*                                                                           */
/*  Debug print of time set ts.  This function accepts NULL for ts.          */
/*                                                                           */
/*****************************************************************************/

/* *** moved to platform, khe_time_set.c
static void KheTimeSetDebug(KHE_TIME_SET ts, int indent, FILE *fp)
{
  KHE_TIME t;  int i;
  fprintf(fp, "%*s", indent, "");
  if( ts == NULL )
    fprintf(fp, "-");
  else
  {
    fprintf(fp, "{");
    HaArrayForEach(ts->times, t, i)
      fprintf(fp, "%s%s", i > 0 ? ", " : "", KheTimeId(t));
    fprintf(fp, "}");
  }
  fprintf(fp, "\n");
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "group by frame solvers"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_GROUP_BY_FRAME_SOLVER KheGroupByFrameSolverMake(                     */
/*    KHE_TASKING tasking, int max_days, bool resource_invariant,            */
/*    KHE_FRAME common_frame, KHE_EVENT_TIMETABLE_MONITOR etm)               */
/*                                                                           */
/*  Make a new group by frame solver object.                                 */
/*                                                                           */
/*****************************************************************************/

static KHE_GROUP_BY_FRAME_SOLVER KheGroupByFrameSolverMake(
  KHE_TASKING tasking, int max_days, bool resource_invariant,
  KHE_FRAME common_frame, KHE_EVENT_TIMETABLE_MONITOR etm,
  KHE_TASK_SET ts, HA_ARENA a)
{
  KHE_GROUP_BY_FRAME_SOLVER res;  int count;  KHE_INSTANCE ins;

  /* the solve as a whole */
  HaMake(res, a);
  res->arena = a;
  res->tasking = tasking;
  res->soln = KheTaskingSoln(tasking);
  res->resource_invariant = resource_invariant;
  res->common_frame = common_frame;
  res->etm = etm;
  res->resource = NULL;
  res->soln_changed = false;
  res->task_set = ts;

  /* phase 0 */
  count = KheFrameTimeGroupCount(common_frame);
  HaArrayInit(res->provisionally_linked_to_prev, a);
  HaArrayFill(res->provisionally_linked_to_prev, count, false);
  HaArrayInit(res->linked_to_prev, a);
  HaArrayFill(res->linked_to_prev, count, false);

  /* phase 1 */
  res->max_days = max_days;
  HaArrayInit(res->framed_monitors, a);
  HaArrayInit(res->start_indexes, a);
  res->group_monitor = NULL;
  ins = KheSolnInstance(res->soln);
  res->successful_times = KheTimeSetMake(ins, a);
  res->last_index = 0;
  res->success_count = 0;
  res->success_ts = KheTimeSetMake(ins, a);
  res->curr_ts = KheTimeSetMake(ins, a);

  /* phase 2 */
  HaArrayInit(res->time_sets, a);
  HaArrayInit(res->time_sets_by_time, a);
  HaArrayInit(res->time_sets_by_monitor, a);
  HaArrayInit(res->suitable_monitors, a);
  HaArrayInit(res->success_seq, a);
  HaArrayInit(res->curr_seq, a);
  HaArrayInit(res->task_sets, a);
  HaArrayInit(res->counts, a);
  HaArrayInit(res->domain_task_sets, a);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheTaskCmp(const void *t1, const void *t2)                           */
/*                                                                           */
/*  Comparison function for sorting tasks so that they can be uniqueified.   */
/*                                                                           */
/*****************************************************************************/

static int KheTaskCmp(const void *t1, const void *t2)
{
  KHE_TASK task1 = * (KHE_TASK *) t1;
  KHE_TASK task2 = * (KHE_TASK *) t2;
  return KheTaskSolnIndex(task1) - KheTaskSolnIndex(task2);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_TASK_SET KheFindAllTasksAtTime(KHE_GROUP_BY_FRAME_SOLVER gfs,        */
/*    KHE_TIME time)                                                         */
/*                                                                           */
/*  Return a new task set containing all the tasks of duration 1 running     */
/*  at time, or rather, their proper roots, uniqueified.  Only tasks with    */
/*  the appropriate type, and only tasks which would incur a cost if not     */
/*  assigned, are included.                                                  */
/*                                                                           */
/*****************************************************************************/

static KHE_TASK_SET KheFindAllTasksAtTime(KHE_GROUP_BY_FRAME_SOLVER gfs,
  KHE_TIME time)
{
  KHE_TASK_SET res;  KHE_MEET meet;  KHE_TASK task;  int i, j;
  KHE_RESOURCE_TYPE rt;
  rt = KheResourceResourceType(gfs->resource);
  res = KheTaskSetMake(gfs->soln);
  for( i = 0; i < KheEventTimetableMonitorTimeMeetCount(gfs->etm, time); i++ )
  {
    meet = KheEventTimetableMonitorTimeMeet(gfs->etm, time, i);
    if( KheMeetDuration(meet) == 1 )
      for( j = 0;  j < KheMeetTaskCount(meet);  j++ )
      {
	task = KheTaskProperRoot(KheMeetTask(meet, j));
	if( KheTaskResourceType(task)==rt && KheTaskNonAssignmentHasCost(task) )
	  KheTaskSetAddTask(res, task);
      }
  }
  KheTaskSetSortUnique(res, &KheTaskCmp);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheFindAndAssignTaskAtTime(KHE_GROUP_BY_FRAME_SOLVER gfs,           */
/*    KHE_TIME time, KHE_TASK *task)                                         */
/*                                                                           */
/*  Find and assign an arbitrary unassigned task of duration 1 at this       */
/*  time, and return true if successful.                                     */
/*                                                                           */
/*****************************************************************************/

static bool KheFindAndAssignTaskAtTime(KHE_GROUP_BY_FRAME_SOLVER gfs,
  KHE_TIME time, KHE_TASK *task)
{
  KHE_MEET meet;  KHE_TASK task2;  int i, j;
  for( i = 0; i < KheEventTimetableMonitorTimeMeetCount(gfs->etm, time); i++ )
  {
    meet = KheEventTimetableMonitorTimeMeet(gfs->etm, time, i);
    if( KheMeetDuration(meet) == 1 )
      for( j = 0;  j < KheMeetTaskCount(meet);  j++ )
      {
	task2 = KheTaskProperRoot(KheMeetTask(meet, j));
	if( KheTaskAssignResource(task2, gfs->resource) )
	  return *task = task2, true;
      }
  }
  return *task = NULL, false;
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "phase 0"                                                      */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  int IntCmpIncreasing(const void *t1, const void *t2)                     */
/*                                                                           */
/*  Comparison fuunction for sorting an array of integers into increasing    */
/*  order.                                                                   */
/*                                                                           */
/*****************************************************************************/

static int IntCmpIncreasing(const void *t1, const void *t2)
{
  int val1 = * (int *) t1;
  int val2 = * (int *) t2;
  return val1 - val2;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDemandAtTime(KHE_EVENT_TIMETABLE_MONITOR etm,                     */
/*    KHE_RESOURCE_TYPE rt, KHE_TIME t)                                      */
/*                                                                           */
/*  Return the total demand for resources of type rt at time t.              */
/*                                                                           */
/*****************************************************************************/

static int KheDemandAtTime(KHE_EVENT_TIMETABLE_MONITOR etm,
  KHE_RESOURCE_TYPE rt, KHE_TIME t)
{
  int j, k, res;  KHE_MEET meet;  KHE_TASK task;  KHE_RESOURCE r;
  res = 0;
  for( j = 0;  j < KheEventTimetableMonitorTimeMeetCount(etm, t);  j++ )
  {
    meet = KheEventTimetableMonitorTimeMeet(etm, t, j);
    for( k = 0;  k < KheMeetTaskCount(meet);  k++ )
    {
      task = KheMeetTask(meet, k);
      if( KheTaskResourceType(task) == rt &&
	(KheTaskIsPreassigned(task, &r) || KheTaskNonAssignmentHasCost(task)) )
	res++;
    }
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDemandAtTimeGroup(KHE_EVENT_TIMETABLE_MONITOR etm,                */
/*    KHE_RESOURCE_TYPE rt, KHE_TIME_GROUP tg)                               */
/*                                                                           */
/*  Return the total demand for resources of type rt in the times of tg.     */
/*                                                                           */
/*****************************************************************************/

static int KheDemandAtTimeGroup(KHE_EVENT_TIMETABLE_MONITOR etm,
  KHE_RESOURCE_TYPE rt, KHE_TIME_GROUP tg)
{
  int i, res;  KHE_TIME t;
  res = 0;
  for( i = 0;  i < KheTimeGroupTimeCount(tg);  i++ )
  {
    t = KheTimeGroupTime(tg, i);
    res += KheDemandAtTime(etm, rt, t);
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDemandAtClusterMonitor(KHE_EVENT_TIMETABLE_MONITOR etm,           */
/*    KHE_RESOURCE_TYPE rt, KHE_CLUSTER_BUSY_TIMES_MONITOR cbtm)             */
/*                                                                           */
/*  Return the total demand for resources of type rt in the time groups      */
/*  of cluster busy times monitor cbtm.                                      */
/*                                                                           */
/*****************************************************************************/

static int KheDemandAtClusterMonitor(KHE_EVENT_TIMETABLE_MONITOR etm,
  KHE_RESOURCE_TYPE rt, KHE_CLUSTER_BUSY_TIMES_MONITOR cbtm)
{
  int i, res;  KHE_TIME_GROUP tg;  KHE_POLARITY po;
  res = 0;
  for( i = 0;  i < KheClusterBusyTimesMonitorTimeGroupCount(cbtm);  i++ )
  {
    tg = KheClusterBusyTimesMonitorTimeGroup(cbtm, i, &po);
    res += KheDemandAtTimeGroup(etm, rt, tg);
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KhePhaseZeroForMonitor(KHE_GROUP_BY_FRAME_SOLVER gfs,               */
/*    KHE_FRAMED_MONITOR fm, int lim)                                        */
/*                                                                           */
/*  Apply Phase 0 to fm, which is a suitable cluster busy times monitor      */
/*  with the given lim.                                                      */
/*                                                                           */
/*****************************************************************************/

static void KhePhaseZeroForMonitor(KHE_GROUP_BY_FRAME_SOLVER gfs,
  KHE_FRAMED_MONITOR fm, int lim)
{
  KHE_CLUSTER_BUSY_TIMES_MONITOR cbtm;  KHE_TIME_GROUP m_tg, f_tg, prev_f_tg;
  int i, j, count, sum;  KHE_POLARITY po;  bool val;
  bool f_tg_overlaps, prev_f_tg_overlaps;
  if( DEBUG6 )
  {
    fprintf(stderr, "[ KhePhaseZeroForMonitor(gfs, fm, %d), cbtm =\n", lim);
    KheMonitorDebug(fm->monitor, 3, 2, stderr);
  }

  /* for each time group of fm, find the number of frame time groups it
     intersects with */
  fm->phase0_lim = lim;
  cbtm = (KHE_CLUSTER_BUSY_TIMES_MONITOR) fm->monitor;
  for( i = 0;  i < HaArrayCount(gfs->provisionally_linked_to_prev);  i++ )
    HaArrayPut(gfs->provisionally_linked_to_prev, i, false);
  for( i = 0;  i < KheClusterBusyTimesMonitorTimeGroupCount(cbtm);  i++ )
  {
    m_tg = KheClusterBusyTimesMonitorTimeGroup(cbtm, i, &po);
    count = 0;
    prev_f_tg = NULL;
    prev_f_tg_overlaps = false;
    for( j = fm->first_index;  j <= fm->last_index;  j++ )
    {
      f_tg = KheFrameTimeGroup(gfs->common_frame, j);
      f_tg_overlaps = !KheTimeGroupDisjoint(m_tg, f_tg);
      if( f_tg_overlaps )
	count++;
      if( f_tg_overlaps && prev_f_tg_overlaps )
      {
	/* if we later find that D(T) >= S(T), then if prev_f_tg is */
	/* busy, then f_tg must be busy as well, because they both  */
	/* overlap the same time group of the monitor */
	if( DEBUG6 )
	{
	  fprintf(stderr, "  provisionally linked time groups ");
	  KheTimeGroupDebug(prev_f_tg, 1, -1, stderr);
	  fprintf(stderr, " and ");
	  KheTimeGroupDebug(f_tg, 1, -1, stderr);
	  fprintf(stderr, "\n");
	}

	/* record the linkage provisionally */
	HaArrayPut(gfs->provisionally_linked_to_prev, j, true);
      }
      prev_f_tg = f_tg;
      prev_f_tg_overlaps = f_tg_overlaps;
    }
    HaArrayAddLast(fm->phase0_intersects, count);
  }

  /* sort the intersect counts into increasing order and sum the last lim */
  HaArraySort(fm->phase0_intersects, &IntCmpIncreasing);
  sum = 0;
  count = HaArrayCount(fm->phase0_intersects);
  for( i = 1;  i <= lim;  i++ )
    sum += HaArray(fm->phase0_intersects, count - i);
  count = KheResourceTypeResourceCount(KheResourceResourceType(gfs->resource));
  fm->phase0_supply = sum * count;

  /* get the demand */
  fm->phase0_demand = KheDemandAtClusterMonitor(gfs->etm,
    KheResourceResourceType(gfs->resource), cbtm);

  /* if demand >= supply, provisional linkages become definite */
  if( fm->phase0_demand >= fm->phase0_supply )
  {
    HaArrayForEach(gfs->provisionally_linked_to_prev, val, i)
      if( val )
	HaArrayPut(gfs->linked_to_prev, i, true);
  }

  if( DEBUG6 )
  {
    KheFramedMonitorDebug(fm, 2, stderr);
    fprintf(stderr, "] KhePhaseZeroForMonitor returning\n");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheClusterConstraintHasNonSingletonTimeGroup(                       */
/*    KHE_CLUSTER_BUSY_TIMES_CONSTRAINT cbtc)                                */
/*                                                                           */
/*  Return true if cbtc has at least one time group with >=2 elements.       */
/*                                                                           */
/*****************************************************************************/

static bool KheClusterConstraintHasNonSingletonTimeGroup(
  KHE_CLUSTER_BUSY_TIMES_CONSTRAINT cbtc)
{
  int i;  KHE_TIME_GROUP tg;  KHE_POLARITY po;
  for( i = 0;  i < KheClusterBusyTimesConstraintTimeGroupCount(cbtc);  i++ )
  {
    tg = KheClusterBusyTimesConstraintTimeGroup(cbtc, i, 0, &po);
    if( KheTimeGroupTimeCount(tg) >= 2 )
      return true;
  }
  return false;
}


/*****************************************************************************/
/*                                                                           */
/*  static bool KheClusterConstraintSuitsPhase0(                             */
/*    KHE_CLUSTER_BUSY_TIMES_CONSTRAINT cbtc, int *lim)                      */
/*                                                                           */
/*  Return true if c is suitable for Phase 0, that is, if its time groups    */
/*  are all positive and it has a non-trivial maximum limit, or if its       */
/*  time groups are all negative and it has a non-trivial minimum limit.     */
/*                                                                           */
/*  Set *lim to a maximum limit on the number of busy (not active) time      */
/*  groups in c.  This will be c's maximum limit if its time groups are      */
/*  all positive, and the number of time groups minus the minimum limit if   */
/*  its time groups are all negative.                                        */
/*                                                                           */
/*****************************************************************************/

static bool KheClusterConstraintSuitsPhase0(
  KHE_CLUSTER_BUSY_TIMES_CONSTRAINT cbtc, int *lim)
{
  int max_lim, min_lim;
  max_lim = KheClusterBusyTimesConstraintMaximum(cbtc);
  min_lim = KheClusterBusyTimesConstraintMinimum(cbtc);
  if( KheClusterBusyTimesConstraintAllPositive(cbtc) &&
      max_lim < KheClusterBusyTimesConstraintTimeGroupCount(cbtc) )
  {
    *lim = max_lim;
    return true;
  }
  else if( KheClusterBusyTimesConstraintAllNegative(cbtc) && min_lim > 0 )
  {
    *lim = KheClusterBusyTimesConstraintTimeGroupCount(cbtc) - min_lim;
    return true;
  }
  else
  {
    *lim = -1;
    return false;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheGroupByFramePhase0(KHE_GROUP_BY_FRAME_SOLVER gfs)                */
/*                                                                           */
/*  Carry out Phase 0 of group by frame.                                     */
/*                                                                           */
/*  NB the condition ClusterBusyTimesConstraintTimeGroupsDisjoint(cbtc) is   */
/*  not really required, but it always seems to hold, and it makes the       */
/*  calculation of demand much easier.                                       */
/*                                                                           */
/*****************************************************************************/
static void KheGroupByFrameSolverDebug(KHE_GROUP_BY_FRAME_SOLVER gfs,
  int indent, FILE *fp);

static void KheGroupByFramePhase0(KHE_GROUP_BY_FRAME_SOLVER gfs)
{
  KHE_FRAMED_MONITOR fm;  int i, lim;
  KHE_CLUSTER_BUSY_TIMES_MONITOR cbtm;
  KHE_CLUSTER_BUSY_TIMES_CONSTRAINT cbtc;
  if( DEBUG6 )
    fprintf(stderr, "[ KheGroupByFramePhase0(%s)\n",
      KheResourceTypeId(KheResourceResourceType(gfs->resource)));
  for( i = 0;  i < HaArrayCount(gfs->linked_to_prev);  i++ )
    HaArrayPut(gfs->linked_to_prev, i, false);
  HaArrayForEach(gfs->framed_monitors, fm, i)
    if( KheMonitorTag(fm->monitor) == KHE_CLUSTER_BUSY_TIMES_MONITOR_TAG )
    {
      cbtm = (KHE_CLUSTER_BUSY_TIMES_MONITOR) fm->monitor;
      cbtc = KheClusterBusyTimesMonitorConstraint(cbtm);
      if( ClusterBusyTimesConstraintTimeGroupsDisjoint(cbtc) &&
	 KheClusterConstraintSuitsPhase0(cbtc, &lim) &&
         KheClusterConstraintHasNonSingletonTimeGroup(cbtc) )
	KhePhaseZeroForMonitor(gfs, fm, lim);
    }
  if( DEBUG6 )
  {
    fprintf(stderr, "  at end of KheGroupByFramePhase0:\n");
    KheGroupByFrameSolverDebug(gfs, 2, stderr);
    fprintf(stderr, "] KheGroupByFramePhase0 returning\n");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "phase 1"                                                      */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheGroupByFrameAssign(KHE_GROUP_BY_FRAME_SOLVER gfs,                */
/*    int curr_index, bool prev_tg_assigned)                                 */
/*                                                                           */
/*  Carry out the part of the solve that assigns all combinations of tasks   */
/*  and checks which ones are successful.                                    */
/*                                                                           */
/*  If prev_tg_assigned, then the previous time group (at curr_index - 1)    */
/*  has been assigned something.                                             */
/*                                                                           */
/*****************************************************************************/

static void KheGroupByFrameAssign(KHE_GROUP_BY_FRAME_SOLVER gfs,
  int curr_index, bool prev_tg_assigned)
{
  KHE_TIME_GROUP tg;  int i;  KHE_TIME t;  KHE_TASK task;
  if( curr_index > gfs->last_index )
  {
    /* curr_ts is complete; record it if successful (has zero cost) */
    if( KheMonitorCost((KHE_MONITOR) gfs->group_monitor) == 0 )
    {
      /* curr_ts is a success */
      gfs->success_count++;
      KheTimeSetClear(gfs->success_ts);
      KheTimeSetUnion(gfs->success_ts, gfs->curr_ts);
      /* KheTimeSetCopy(gfs->success_ts, gfs->curr_ts); */
    }
  }
  else
  {
    /* try assigning at time in the current time group */
    tg = KheFrameTimeGroup(gfs->common_frame, curr_index);
    for( i = 0;  gfs->success_count < 2 && i < KheTimeGroupTimeCount(tg);  i++ )
    {
      t = KheTimeGroupTime(tg, i);
      if( KheFindAndAssignTaskAtTime(gfs, t, &task) )
      {
	KheTimeSetAddTime(gfs->curr_ts, t);
        KheGroupByFrameAssign(gfs, curr_index + 1, true);
	KheTimeSetDeleteLastTime(gfs->curr_ts);
	/* HaArrayDeleteLast(gfs->curr_ts->times); */
	KheTaskUnAssign(task);
      }
    }

    /* try not assigning within the current time group */
    if( prev_tg_assigned && HaArray(gfs->linked_to_prev, curr_index) )
    {
      if( DEBUG6 )
      {
	fprintf(stderr, "  due to linkage, omitting non-assignment of ");
	KheTimeGroupDebug(tg, 1, 0, stderr);
      }
    }
    else
      KheGroupByFrameAssign(gfs, curr_index + 1, false);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheGroupByFrameGroupTasks(KHE_GROUP_BY_FRAME_SOLVER gfs,            */
/*     KHE_TIME_SET ts)                                                      */
/*                                                                           */
/*  Group the tasks identified by ts, and set gfs->soln_changed if some      */
/*  actual grouping occurs.                                                  */
/*                                                                           */
/*****************************************************************************/

static void KheGroupByFrameGroupTasks(KHE_GROUP_BY_FRAME_SOLVER gfs,
   KHE_TIME_SET ts)
{
  KHE_TASK_SET task_set;  KHE_TIME t;
  KHE_TASK first_task, task;  int i, j, count, min_count;  bool ok;
  if( KheTimeSetTimeCount(ts) >= 2 )
  {
    /* get one task set for each time; also calculate the smallest size */
    HaArrayClear(gfs->task_sets);
    min_count = INT_MAX;
    /* HaArrayForEach(ts->times, t, i) */
    for( i = 0;  i < KheTimeSetTimeCount(ts);  i++ )
    {
      t = KheTimeSetTime(ts, i);
      task_set = KheFindAllTasksAtTime(gfs, t);
      HaArrayAddLast(gfs->task_sets, task_set);
      count = KheTaskSetTaskCount(task_set);
      if( count < min_count )
	min_count = count;
    }

    /* carry out the groupings; NB none of the tasks are currently assigned */
    for( i = 0;  i < min_count;  i++ )
    {
      /* group together the i'th tasks from each task set */
      task_set = HaArray(gfs->task_sets, 0);
      first_task = KheTaskSetTask(task_set, i);
      if( DEBUG3 )
      {
	fprintf(stderr, "    KheGroupByFrame grouping tasks {");
	KheTaskDebug(first_task, 1, -1, stderr);
      }
      for( j = 1;  j < HaArrayCount(gfs->task_sets);  j++ )
      {
	task_set = HaArray(gfs->task_sets, j);
	task = KheTaskSetTask(task_set, i);
	if( task == first_task )
	  ok = true;
	else
	{
	  ok = KheTaskAssign(task, first_task);
	  if( ok )
	  {
	    gfs->soln_changed = true;
	    if( gfs->task_set != NULL )
	      KheTaskSetAddTask(gfs->task_set, task);
	  }
	}
	if( DEBUG3 )
	{
	  fprintf(stderr, ", %s", ok ? "" : "not ");
	  KheTaskDebug(task, 1, -1, stderr);
	}
      }
      if( DEBUG3 )
	fprintf(stderr, "}\n");
    }

    /* free the task sets */
    HaArrayForEach(gfs->task_sets, task_set, i)
      KheTaskSetDelete(task_set);
    HaArrayClear(gfs->task_sets);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheGroupByFramePhase1ForInterval(KHE_GROUP_BY_FRAME_SOLVER gfs,     */
/*    int first_index, int last_index)                                       */
/*                                                                           */
/*  Solve gfs for the interval first_index .. last_index, and return         */
/*  true if there was exactly one success, with gfs->success_ts holding      */
/*  the successful set of times.                                             */
/*                                                                           */
/*****************************************************************************/

static void KheGroupByFramePhase1ForInterval(KHE_GROUP_BY_FRAME_SOLVER gfs,
  int first_index, int last_index)
{
  int i, si;  bool success;  KHE_FRAMED_MONITOR fm;  KHE_TIME_GROUP tg;
  KHE_TIME t;  KHE_TASK task;  KHE_TIME_SET ts;  KHE_INSTANCE ins;
  if( DEBUG2 )
    fprintf(stderr,
      "  [ KheGroupByFrameSolverFindAsst(gfs, first %d, last %d)\n",
      first_index, last_index);
  ins = KheSolnInstance(gfs->soln);

  /* gather relevant monitors into a group monitor */
  gfs->group_monitor = KheGroupMonitorMake(gfs->soln, 1, "GroupByFrame");
  si = (first_index < HaArrayCount(gfs->start_indexes) ?
    HaArray(gfs->start_indexes, first_index) :
    HaArrayCount(gfs->framed_monitors));
  for( i = si;  i < HaArrayCount(gfs->framed_monitors);  i++ )
  {
    /* fm starts at or after first_index */
    fm = HaArray(gfs->framed_monitors, i);
    HnAssert(fm->first_index >= first_index,
      "KheGroupByFrameSolverFindAsst internal error");
    if( fm->last_index <= last_index )
    {
      KheGroupMonitorAddChildMonitor(gfs->group_monitor, fm->monitor);
      if( DEBUG2 )
      {
	fprintf(stderr, "    adding ");
	KheFramedMonitorDebug(fm, 0, stderr);
      }
    }

    /* if fm->first_index > last_index, there is no point in going further */
    if( fm->first_index > last_index )
      break;
  }

  /* try one solve for each time in the first time group, but then */
  /* start all over again and have another go if any were successful */
  tg = KheFrameTimeGroup(gfs->common_frame, first_index);
  KheTimeSetClear(gfs->successful_times);
  do
  {
    success = false;
    for( i = 0;  i < KheTimeGroupTimeCount(tg);  i++ )
    {
      /* initialize the solve fields other than gm */
      gfs->last_index = last_index;
      gfs->success_count = 0;
      KheTimeSetClear(gfs->success_ts);
      KheTimeSetClear(gfs->curr_ts);

      /* solve for t, unless already successful */
      t = KheTimeGroupTime(tg, i);
      if( !KheTimeSetContainsTime(gfs->successful_times, t) )
      {
	if( KheFindAndAssignTaskAtTime(gfs, t, &task) )
	{
	  KheTimeSetAddTime(gfs->curr_ts, t);
	  KheGroupByFrameAssign(gfs, first_index + 1, true);
	  KheTimeSetDeleteLastTime(gfs->curr_ts);
	  /* HaArrayDeleteLast(gfs->curr_ts->times); */
	  KheTaskUnAssign(task);
	}

	/* if exactly one success, we have succeeded */
	if( gfs->success_count == 1 )
	{
	  /* have a single successful assignment, so group tasks */
	  if( DEBUG2 )
	  {
	    fprintf(stderr, "  KheGroupByFrameSolverFindAsst found group: ");
	    KheTimeSetDebug(gfs->success_ts, 2, 0, stderr);
	  }
	  ts = KheTimeSetMake(ins, gfs->arena);
	  KheTimeSetUnion(ts, gfs->success_ts);
	  /* KheTimeSetCopy(ts, gfs->success_ts); */
	  HaArrayAddLast(gfs->time_sets, ts);
	  KheGroupByFrameGroupTasks(gfs, ts);
          KheTimeSetAddTime(gfs->successful_times, t);
	  success = true;
	}
      }
    }
  } while( success );

  /* clear out for next (rt, f), and return */
  KheGroupMonitorDelete(gfs->group_monitor);
  gfs->group_monitor = NULL;
  if( DEBUG2 )
    fprintf(stderr, "  ] KheGroupByFrameSolverFindAsst returning\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheGroupByFrameSolverDebug(KHE_GROUP_BY_FRAME_SOLVER gfs,           */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Produce a debug print of gfs on fp with the given indent.                */
/*                                                                           */
/*****************************************************************************/

static void KheGroupByFrameSolverDebug(KHE_GROUP_BY_FRAME_SOLVER gfs,
  int indent, FILE *fp)
{
  KHE_FRAMED_MONITOR fm;  int i, pos;  KHE_TIME_GROUP tg;  bool val;
  fprintf(fp, "%*s[ GroupByFrameSolver(%s, r %s, max_days %d, rs_invt %s)\n", 
    indent, "", KheInstanceId(KheSolnInstance(gfs->soln)),
    KheResourceId(gfs->resource), gfs->max_days,
    gfs->resource_invariant ? "true" : "false");
  fprintf(fp, "  [ common frame (# indicates linked to previous):\n");
  for( i = 0;  i < KheFrameTimeGroupCount(gfs->common_frame);  i++ )
  {
    tg = KheFrameTimeGroup(gfs->common_frame, i);
    val = HaArray(gfs->linked_to_prev, i);
    fprintf(fp, "    %c ", val ? '#' : ' ');
    KheTimeGroupDebug(tg, 1, 0, fp);
  }
  fprintf(fp, "  ]\n");
  HaArrayForEach(gfs->framed_monitors, fm, i)
  {
    fprintf(fp, "%*s%2d: ", indent, "", i);
    KheFramedMonitorDebug(fm, 0, fp);
  }
  fprintf(fp, "%*s[", indent + 2, "");
  HaArrayForEach(gfs->start_indexes, pos, i)
    fprintf(fp, "%s%d", i > 0 ? ", " : "", pos);
  fprintf(fp, "]\n");
  fprintf(fp, "%*s]\n", indent, "");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheGroupByFramePhase1(KHE_GROUP_BY_FRAME_SOLVER gfs)                */
/*                                                                           */
/*  Carry out group by frame phase 1.  Field gfs->resource is non-NULL       */
/*  and determines which resource type, and which monitors, to use.          */
/*                                                                           */
/*****************************************************************************/

static void KheGroupByFramePhase1(KHE_GROUP_BY_FRAME_SOLVER gfs)
{
  KHE_FRAMED_MONITOR fm;  int i, pos, si, len, first_index, days;
  KHE_MONITOR m;

  /* build up and sort the framed_monitors array */
  HnAssert(HaArrayCount(gfs->framed_monitors) == 0,
    "KheGroupByFramePhase1 internal error 1");
  for( i = 0;  i < KheSolnResourceMonitorCount(gfs->soln, gfs->resource);  i++ )
  {
    m = KheSolnResourceMonitor(gfs->soln, gfs->resource, i);
    if( KheUseMonitor(m, gfs->common_frame, &fm, gfs->arena) )
      HaArrayAddLast(gfs->framed_monitors, fm);
  }
  HaArraySort(gfs->framed_monitors, &KheFramedMonitorCmp);

  /* build up the start_indexes array */
  HnAssert(HaArrayCount(gfs->start_indexes) == 0,
    "KheGroupByFramePhase1 internal error 1");
  pos = 0;
  for( si = 0;  pos < HaArrayCount(gfs->framed_monitors);  si++ )
  {
    /* add the index for si */
    fm = HaArray(gfs->framed_monitors, pos);
    while( fm->first_index < si )
    {
      pos++;
      if( pos >= HaArrayCount(gfs->framed_monitors) )
	break;
      fm = HaArray(gfs->framed_monitors, pos);
    }
    HaArrayAddLast(gfs->start_indexes, pos);
  }

  /* carry out Phase 0 */
  KheGroupByFramePhase0(gfs);

  if( DEBUG2 )
    KheGroupByFrameSolverDebug(gfs, 2, stderr);

  /* solve for each subsequence of up to max_days days */
  days = KheFrameTimeGroupCount(gfs->common_frame);
  for( len = gfs->max_days;  len >= 2;  len-- )
    for( first_index = 0;  first_index <= days - len;  first_index++ )
      KheGroupByFramePhase1ForInterval(gfs, first_index, first_index + len-1);

  /* clear out the framed_monitors and start_indexes arrays */
  /* ***
  HaArrayForEach(gfs->framed_monitors, fm, i)
    KheFramedMonitorFree(fm);
  *** */
  HaArrayClear(gfs->framed_monitors);
  HaArrayClear(gfs->start_indexes);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "phase 2"                                                      */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  bool KheLimitActiveIntervalsMonitorIsSuitable(                           */
/*    KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim)                               */
/*                                                                           */
/*  Return true if laim is suitable for Phase 2:  if its constraint applies  */
/*  to all resources, and all its time groups are singletons and positive.   */
/*                                                                           */
/*****************************************************************************/

static bool KheLimitActiveIntervalsMonitorIsSuitable(
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim)
{
  int i;  KHE_TIME_GROUP tg;  KHE_POLARITY po;  KHE_RESOURCE_GROUP rg;
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT laic;

  /* if monitor's constraint does not monitor all resources, return false */
  laic = KheLimitActiveIntervalsMonitorConstraint(laim);
  if( KheLimitActiveIntervalsConstraintResourceGroupCount(laic) != 1 )
    return false;
  rg = KheLimitActiveIntervalsConstraintResourceGroup(laic, 0);
  if( KheResourceGroupResourceCount(rg) !=
      KheResourceTypeResourceCount(KheResourceGroupResourceType(rg)) )
    return false;

  /* there must be at least two time groups */
  if( KheLimitActiveIntervalsMonitorTimeGroupCount(laim) < 2 )
    return false;

  /* if one of its time groups is not single or not positive, return false */
  for( i = 0;  i < KheLimitActiveIntervalsMonitorTimeGroupCount(laim);  i++ )
  {
    tg = KheLimitActiveIntervalsMonitorTimeGroup(laim, i, &po);
    if( KheTimeGroupTimeCount(tg) != 1 || po != KHE_POSITIVE )
      return false;
  }

  /* all OK */
  return true;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheLimitActiveIntervalsMonitorsCanGroupTogether(                    */
/*    KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim1,                              */
/*    KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim2)                              */
/*                                                                           */
/*  Return true if laim1 and laim2 have equal time groups.                   */
/*                                                                           */
/*****************************************************************************/

static bool KheLimitActiveIntervalsMonitorsCanGroupTogether(
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim1,
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim2)
{
  int i;  KHE_TIME_GROUP tg1, tg2;  KHE_POLARITY po1, po2;
  bool has_min1, has_min2, has_max1, has_max2;

  /* time groups must be equal */
  if( KheLimitActiveIntervalsMonitorTimeGroupCount(laim1) !=
      KheLimitActiveIntervalsMonitorTimeGroupCount(laim2) )
    return false;
  for( i = 0;  i < KheLimitActiveIntervalsMonitorTimeGroupCount(laim1);  i++ )
  {
    tg1 = KheLimitActiveIntervalsMonitorTimeGroup(laim1, i, &po1);
    tg2 = KheLimitActiveIntervalsMonitorTimeGroup(laim2, i, &po2);
    if( !KheTimeGroupEqual(tg1, tg2) )
      return false;
  }

  /* one must have a min limit only, the other a max limit only */
  has_min1 = KheLimitActiveIntervalsMonitorMinimum(laim1) > 0;
  has_min2 = KheLimitActiveIntervalsMonitorMinimum(laim2) > 0;
  has_max1 = KheLimitActiveIntervalsMonitorMaximum(laim1) <
    KheLimitActiveIntervalsMonitorTimeGroupCount(laim1);
  has_max2 = KheLimitActiveIntervalsMonitorMaximum(laim2) <
    KheLimitActiveIntervalsMonitorTimeGroupCount(laim2);
  if( (has_min1 && !has_max1) && (!has_min2 && has_max2) )
    return true;
  if( (!has_min1 && has_max1) && (has_min2 && !has_max2) )
    return true;
  return false;
}


/*****************************************************************************/
/*                                                                           */
/*  void FindSeq(KHE_GROUP_BY_FRAME_SOLVER gfs, int min_limit,               */
/*    int max_limit, int remaining_len)                                      */
/*                                                                           */
/*  This is the recursive part of finding a unique sequence of               */
/*  sub-interval lengths, if there is one.  Each length must lie             */
/*  between min_limit and max_limit; the remaining wanted length             */
/*  is remaining_len.                                                        */
/*                                                                           */
/*****************************************************************************/

static void FindSeq(KHE_GROUP_BY_FRAME_SOLVER gfs, int min_limit,
  int max_limit, int remaining_len)
{
  int first_len, len, i;
  if( ++(gfs->node_count) > 100 )
  {
    /* give up, search is taking too long */
    if( DEBUG5 )
      fprintf(stderr, "  curr_seq search taking too long\n");
    gfs->success_count = 2;
  }
  else if( remaining_len < 0 )
  {
    /* no solutions here */
  }
  else if( remaining_len == 0 )
  {
    /* curr_seq is a feasible soln; but no good if more than one */
    if( DEBUG5 )
    {
      fprintf(stderr, "  successful curr_seq:");
      HaArrayForEach(gfs->curr_seq, len, i)
	fprintf(stderr, " %d", len);
      fprintf(stderr, "\n");
    }
    gfs->success_count++;
    if( gfs->success_count >= 2 )
      return;

    /* if any of the lengths of curr_seq differ, there is more than one soln */
    first_len = HaArrayFirst(gfs->curr_seq);
    for( i = 1;  i < HaArrayCount(gfs->curr_seq);  i++ )
      if( HaArray(gfs->curr_seq, i) != first_len )
      {
	gfs->success_count = 2;
	return;
      }

    /* save curr_seq; at the moment it is the unique feasible solution */
    HaArrayClear(gfs->success_seq);
    HaArrayForEach(gfs->curr_seq, len, i)
      HaArrayAddLast(gfs->success_seq, len);
  }
  else /* remaining_len > 0 */
  {
    for( len = min_limit;  len <= max_limit && gfs->success_count < 2;  len++ )
    {
      HaArrayAddLast(gfs->curr_seq, len);
      FindSeq(gfs, min_limit, max_limit, remaining_len - len);
      HaArrayDeleteLast(gfs->curr_seq);
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheMergeOneGap(KHE_GROUP_BY_FRAME_SOLVER gfs,                       */
/*    KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim, int min_limit,                */
/*    int max_limit, int gap_start, int gap_stop)                            */
/*                                                                           */
/*  Do merges in the gap from gap_start inclusive to gap_stop exclusive.     */
/*  Return true if successful.                                               */
/*                                                                           */
/*****************************************************************************/

static bool KheMergeOneGap(KHE_GROUP_BY_FRAME_SOLVER gfs,
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim, int min_limit,
  int max_limit, int gap_start, int gap_stop)
{
  int i, j, len, start;  KHE_TIME_GROUP tg;  KHE_POLARITY po;  KHE_TIME_SET ts;
  KHE_INSTANCE ins;

  if( DEBUG8 )
  {
    fprintf(stderr, "  [ KheMergeOneGap(gfs, lim[%d..%d], gap[%d..%d]:\n",
      min_limit, max_limit, gap_start, gap_stop - 1);
    for( i = gap_start;  i < gap_stop;  i++ )
    {
      tg = KheLimitActiveIntervalsMonitorTimeGroup(laim, i, &po);
      KheTimeGroupDebug(tg, 1, 6, stderr);
    }
  }

  ins = KheSolnInstance(gfs->soln);
  gfs->success_count = 0;
  HaArrayClear(gfs->success_seq);
  HaArrayClear(gfs->curr_seq);
  gfs->node_count = 0;
  FindSeq(gfs, min_limit, max_limit, gap_stop - gap_start);
  if( DEBUG8 )
  {
    fprintf(stderr, "      found %d solution%s", gfs->success_count,
      gfs->success_count != 1 ? "s" : "");
    if( gfs->success_count == 1 )
    {
      fprintf(stderr, ":");
      HaArrayForEach(gfs->success_seq, len, i)
	fprintf(stderr, " %d", len);
    }
    fprintf(stderr, "\n");
  }
  if( gfs->success_count == 1 )
  {
    /* merge the sets defined by the successful sequence */
    start = gap_start;
    ts = KheTimeSetMake(ins, gfs->arena);
    HaArrayForEach(gfs->success_seq, len, i)
    {
      KheTimeSetClear(ts);
      for( j = 0;  j < len;  j++ )
      {
	tg = KheLimitActiveIntervalsMonitorTimeGroup(laim, start + j, &po);
	KheTimeSetAddTime(ts, KheTimeGroupTime(tg, 0));
      }
      if( DEBUG8 )
      {
	fprintf(stderr, "      merging time set ");
	KheTimeSetDebug(ts, 2, 0, stderr);
      }
      KheGroupByFrameGroupTasks(gfs, ts);
      start += len;
    }
  }

  if( DEBUG8 )
    fprintf(stderr, "  ] KheMergeOneGap returning %s\n",
      bool_show(gfs->success_count == 1));
  return gfs->success_count == 1;
}


/*****************************************************************************/
/*                                                                           */
/*  void KhePlateauMerge(KHE_GROUP_BY_FRAME_SOLVER gfs,                      */
/*    KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim, int min_limit, int max_limit) */
/*                                                                           */
/*  Try the alternative method of merging within one gap, based on           */
/*  the number of tasks beginning and ending.                                */
/*                                                                           */
/*****************************************************************************/

static void KhePlateauMerge(KHE_GROUP_BY_FRAME_SOLVER gfs,
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim, int min_limit, int max_limit)
{
  int i, start_pos, end_pos;  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT laic;
  KHE_DOMAIN_TASK_SET dts;

  /* boilerplate */
  if( DEBUG7 )
  {
    laic = KheLimitActiveIntervalsMonitorConstraint(laim);
    fprintf(stderr, "  [ KhePlateauMerge(gfs, %s, %d..%d)\n",
      KheConstraintId((KHE_CONSTRAINT) laic), min_limit, max_limit);
  }

  /* build the domain task sets */
  KheMakeDomainTaskSets(gfs, laim, min_limit, max_limit);

  /* for each domain task set, find and remove plateaus */
  HaArrayForEach(gfs->domain_task_sets, dts, i)
  {
    if( DEBUG7 )
      KheDomainTaskSetDebug(dts, 2, 4, stderr);
    while( KheDomainTaskSetFindPlateau(dts, &start_pos, &end_pos) )
    {
      if( DEBUG7 )
	KheDomainTaskSetPlateauDebug(dts, start_pos, end_pos, 2, 6, stderr);
      KheDomainTaskSetGroupAndRemovePlateau(dts, start_pos, end_pos, gfs);
    }
  }

  /* free the domain task sets */
  /* still to do */
  if( DEBUG7 )
    fprintf(stderr, "  ] KhePlateauMerge returning\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheGroupTwoMonitors(KHE_GROUP_BY_FRAME_SOLVER gfs,                  */
/*    KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim1,                              */
/*    KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim2)                              */
/*                                                                           */
/*  Carry out Phase 2 grouping based on laim1 and (optionally) laim2.        */
/*                                                                           */
/*****************************************************************************/

static void KheGroupTwoMonitors(KHE_GROUP_BY_FRAME_SOLVER gfs,
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim1,
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim2)
{
  int i, j, min_limit, max_limit, tg_count, index;
  KHE_TIME_GROUP tg;  KHE_POLARITY po;  KHE_TIME t;
  KHE_TIME_SET ts, prev_ts, foll_ts;
  bool history_before, history_after;
  if( DEBUG9 )
  {
    fprintf(stderr, "[ KheGroupTwoMonitors(gfs, laim1, %s)\n",
      laim2 != NULL ? "laim2" : "-");
    KheMonitorDebug((KHE_MONITOR) laim1, 1, 2, stderr);
    if( laim2 != NULL )
      KheMonitorDebug((KHE_MONITOR) laim2, 1, 2, stderr);
  }

  /* set up the time_sets_by_monitor array */
  if( DEBUG9 )
    fprintf(stderr, "    [ time_sets_by_monitor:\n");
  HaArrayClear(gfs->time_sets_by_monitor);
  tg_count = KheLimitActiveIntervalsMonitorTimeGroupCount(laim1);
  for( i = 0;  i < tg_count;  i++ )
  {
    tg = KheLimitActiveIntervalsMonitorTimeGroup(laim1, i, &po);
    HnAssert(KheTimeGroupTimeCount(tg) == 1 && po == KHE_POSITIVE,
      "KheGroupTwoMonitors internal error");
    t = KheTimeGroupTime(tg, 0);
    ts = HaArray(gfs->time_sets_by_time, KheTimeIndex(t));
    HaArrayAddLast(gfs->time_sets_by_monitor, ts);
    if( DEBUG9 )
    {
      fprintf(stderr, "        %d: ", i);
      KheTimeGroupDebug(tg, 1, -1, stderr);
      fprintf(stderr, " --> ");
      if( ts != NULL )
	KheTimeSetDebug(ts, 2, 0, stderr);
      else
	fprintf(stderr, "NULL\n");
    }
  }
  if( DEBUG9 )
    fprintf(stderr, "    ]\n");

  /* if time sets do not coordinate with time groups, exit early */
  for( i = 0;  i < tg_count;  i++ )
  {
    tg = KheLimitActiveIntervalsMonitorTimeGroup(laim1, i, &po);
    ts = HaArray(gfs->time_sets_by_monitor, i);
    if( ts != NULL )
    {
      for( j = i + 1;  j < tg_count;  j++ )
      {
	if( HaArray(gfs->time_sets_by_monitor, j) != ts )
	  break;
      }
      if( j - i != KheTimeSetTimeCount(ts) )
      {
	if( DEBUG9 )
	{
	  fprintf(stderr, "] KheGroupTwoMonitors returning: uncoordinate ");
	  KheTimeSetDebug(ts, 2, 0, stderr);
	}
	return;
      }
      i = j - 1;
    }
  }

  /* get min_limit, max_limit, history_before, and history_after */
  min_limit = KheLimitActiveIntervalsMonitorMinimum(laim1);
  max_limit = KheLimitActiveIntervalsMonitorMaximum(laim1);
  history_before = KheLimitActiveIntervalsMonitorHistoryBefore(laim1) > 0;
  history_after  = KheLimitActiveIntervalsMonitorHistoryAfter(laim1)  > 0;
  if( laim2 != NULL )
  {
    if( KheLimitActiveIntervalsMonitorMinimum(laim2) > min_limit )
      min_limit = KheLimitActiveIntervalsMonitorMinimum(laim2);
    if( KheLimitActiveIntervalsMonitorMaximum(laim2) < max_limit )
      max_limit = KheLimitActiveIntervalsMonitorMaximum(laim2);
    if( KheLimitActiveIntervalsMonitorHistoryBefore(laim2) > 0 )
      history_before = true;
    if( KheLimitActiveIntervalsMonitorHistoryAfter(laim2)  > 0 )
      history_after = true;
  }

  /* extend the leftmost grouped sequence, if present and too short */
  ts = HaArrayFirst(gfs->time_sets_by_monitor);
  if( ts != NULL && !history_before && KheTimeSetTimeCount(ts) < min_limit )
  {
    for( i = KheTimeSetTimeCount(ts);  i < min_limit && i < tg_count;  i++ )
    {
      index = i;
      if( HaArray(gfs->time_sets_by_monitor, index) != NULL )
	break;  /* we could merge here, but we don't */
      tg = KheLimitActiveIntervalsMonitorTimeGroup(laim1, index, &po);
      t = KheTimeGroupTime(tg, 0);
      HaArrayPut(gfs->time_sets_by_monitor, index, ts);
      KheTimeSetAddTime(ts, t);
    }
    if( i > KheTimeSetTimeCount(ts) )
      KheGroupByFrameGroupTasks(gfs, ts);
  }

  /* extend the rightmost grouped sequence, if present and too short */
  ts = HaArrayLast(gfs->time_sets_by_monitor);
  if( ts != NULL && !history_after && KheTimeSetTimeCount(ts) < min_limit )
  {
    for( i = KheTimeSetTimeCount(ts);  i < min_limit && i < tg_count;  i++ )
    {
      index = (tg_count - 1) - i;
      if( HaArray(gfs->time_sets_by_monitor, index) != NULL )
	break;  /* we could merge here, but we don't */
      tg = KheLimitActiveIntervalsMonitorTimeGroup(laim1, index, &po);
      t = KheTimeGroupTime(tg, 0);
      HaArrayPut(gfs->time_sets_by_monitor, index, ts);
      KheTimeSetAddTime(ts, t);
    }
    if( i > KheTimeSetTimeCount(ts) )
      KheGroupByFrameGroupTasks(gfs, ts);
  }

  /* find all gap intervals */
  prev_ts = NULL;
  for( i = 0;  i < KheLimitActiveIntervalsMonitorTimeGroupCount(laim1);  i++ )
  {
    tg = KheLimitActiveIntervalsMonitorTimeGroup(laim1, i, &po);
    ts = HaArray(gfs->time_sets_by_monitor, i);
    if( ts == NULL )
    {
      /* find gap i .. j - 1 */
      foll_ts = NULL;
      for( j=i+1; j < KheLimitActiveIntervalsMonitorTimeGroupCount(laim1); j++ )
      {
	foll_ts = HaArray(gfs->time_sets_by_monitor, j);
	if( foll_ts != NULL )
	  break;
      }

      /* merge within the gap (if prev and foll not extensible); carry on */
      if( DEBUG9 )
	fprintf(stderr, "  found gap %d .. %d\n", i, j - 1);
      if( (prev_ts == NULL || KheTimeSetTimeCount(prev_ts) >= max_limit) &&
          (foll_ts == NULL || KheTimeSetTimeCount(foll_ts) >= max_limit) )
	KheMergeOneGap(gfs, laim1, min_limit, max_limit, i, j);
      i = j - 1;
    }
    else
      prev_ts = ts;
  }

  /* try task merging based on the number of tasks at each time */
  KhePlateauMerge(gfs, laim1, min_limit, max_limit);

  if( DEBUG9 )
    fprintf(stderr, "] KheGroupTwoMonitors returning\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheGroupByFramePhase2(KHE_GROUP_BY_FRAME_SOLVER gfs)                */
/*                                                                           */
/*  Carry out Phase 2 of group by frame.                                     */
/*                                                                           */
/*****************************************************************************/

static void KheGroupByFramePhase2(KHE_GROUP_BY_FRAME_SOLVER gfs)
{
  KHE_MONITOR m;  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim, laimi, laimj;
  int ti, i, j, time_count;  KHE_TIME_SET ts;  KHE_INSTANCE ins;

  /* sort and merge the time sets from Phase 1 */
  /* still to do - in GPost they are all disjoint anyway */
  /* when we do this, we can combine it with the next step */

  /* set up time_sets_by_time, mapping times to time sets where present */
  HaArrayClear(gfs->time_sets_by_time);
  ins = KheSolnInstance(gfs->soln);
  time_count = KheInstanceTimeCount(ins);
  HaArrayFill(gfs->time_sets_by_time, time_count, NULL);
  HaArrayForEach(gfs->time_sets, ts, i)
  {
    /* HaArrayForEach(ts->times, t, j) */
    for( j = 0;  j < KheTimeSetTimeCount(ts);  j++ )
    {
      ti = KheTimeSetTimeIndex(ts, j);
      HaArrayPut(gfs->time_sets_by_time, ti, ts);
    }
  }
  if( DEBUG4 )
  {
    fprintf(stderr, "  [ time_sets_by_time:\n");
    HaArrayForEach(gfs->time_sets_by_time, ts, i)
    {
      fprintf(stderr, "    %d (%s): ", i, KheTimeId(KheInstanceTime(ins, i)));
      if( ts != NULL )
	KheTimeSetDebug(ts, 2, 0, stderr);
      else
	fprintf(stderr, "NULL\n");
    }
    fprintf(stderr, "  ]\n");
  }

  /* find the suitable limit active intervals monitors */
  HaArrayClear(gfs->suitable_monitors);
  for( i = 0;  i < KheSolnResourceMonitorCount(gfs->soln, gfs->resource);  i++ )
  {
    m = KheSolnResourceMonitor(gfs->soln, gfs->resource, i);
    if( KheMonitorTag(m) == KHE_LIMIT_ACTIVE_INTERVALS_MONITOR_TAG )
    {
      laim = (KHE_LIMIT_ACTIVE_INTERVALS_MONITOR) m;
      if( KheLimitActiveIntervalsMonitorIsSuitable(laim) )
      {
	if( DEBUG7 )
	{
	  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c;
	  c = KheLimitActiveIntervalsMonitorConstraint(laim);
	  fprintf(stderr, "  found suitable monitor, constraint %s\n",
	    KheConstraintId((KHE_CONSTRAINT) c));
	}
	HaArrayAddLast(gfs->suitable_monitors, laim);
      }
    }
  }

  /* find pairs of suitable monitors and do the job for each distinct pair */
  for( i = 0;  i < HaArrayCount(gfs->suitable_monitors);  i++ )
  {
    laimi = HaArray(gfs->suitable_monitors, i);
    for( j = i + 1;  j < HaArrayCount(gfs->suitable_monitors);  j++ )
    {
      laimj = HaArray(gfs->suitable_monitors, j);
      if( KheLimitActiveIntervalsMonitorsCanGroupTogether(laimi, laimj) )
      {
	KheGroupTwoMonitors(gfs, laimi, laimj);
	HaArrayDeleteAndPlug(gfs->suitable_monitors, j);  /* order matters! */
	HaArrayDeleteAndPlug(gfs->suitable_monitors, i);  /* order matters! */
	i--;
	break;
      }
    }
  }

  /* find single suitable monitors and do the job for them */
  HaArrayForEach(gfs->suitable_monitors, laim, i)
    KheGroupTwoMonitors(gfs, laim, NULL);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "main function"                                                */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheGroupByFrameForResourceType(KHE_GROUP_BY_FRAME_SOLVER gfs,       */
/*    KHE_RESOURCE_TYPE rt)                                                  */
/*                                                                           */
/*  Carry out group by frame for resource type rt.  There is no point        */
/*  in doing anything if there are no resources, or only one resource.       */
/*                                                                           */
/*****************************************************************************/

static void KheGroupByFrameForResourceType(KHE_GROUP_BY_FRAME_SOLVER gfs,
  KHE_RESOURCE_TYPE rt)
{
  if( KheResourceTypeResourceCount(rt) > 1 )
  {
    gfs->resource = KheResourceTypeResource(rt, 0);
    KheGroupByFramePhase1(gfs);
    KheGroupByFramePhase2(gfs);
    gfs->resource = NULL;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheTaskingGroupByFrame(KHE_TASKING tasking, KHE_OPTIONS options,    */
/*    KHE_TASK_SET ts)                                                       */
/*                                                                           */
/*  Group the tasks of tasking by frame.                                     */
/*                                                                           */
/*  If ts is non-NULL, add every task that is given an assignment to ts.     */
/*                                                                           */
/*****************************************************************************/

bool KheTaskingGroupByFrame(KHE_TASKING tasking, KHE_OPTIONS options,
  KHE_TASK_SET ts)
{
  int max_days, i;  KHE_RESOURCE_TYPE rt;  bool res, resource_invariant;
  KHE_SOLN soln;  KHE_FRAME common_frame;  KHE_EVENT_TIMETABLE_MONITOR etm;
  KHE_INSTANCE ins;  KHE_GROUP_BY_FRAME_SOLVER gfs;  HA_ARENA a;

  /* return immediately if option rs_group_by_frame_off is true */
  soln = KheTaskingSoln(tasking);
  a = KheSolnArenaBegin(soln);
  ins = KheSolnInstance(soln);
  rt = KheTaskingResourceType(tasking);
  if( DEBUG1 )
    fprintf(stderr, "[ KheTaskingGroupByFrame(%s %s, -)\n", KheInstanceId(ins),
      rt == NULL ? "all resource types" : KheResourceTypeId(rt));
  if( KheOptionsGetBool(options, "rs_group_by_frame_off", false) )
  {
    if( DEBUG1 )
      fprintf(stderr, "] KheTaskingGroupByFrame returning false (off)\n");
    return false;
  }

  /* get the other options used by this function */
  max_days = KheOptionsGetInt(options, "rs_group_by_frame_max_days", 3);
  resource_invariant = KheOptionsGetBool(options, "rs_invariant", false);
  etm = (KHE_EVENT_TIMETABLE_MONITOR)
    KheOptionsGetObject(options, "gs_event_timetable_monitor", NULL);
  common_frame = KheOptionsFrame(options, "gs_common_frame", soln);
  HnAssert(etm != NULL, "KheTaskingGroupByFrame: required option "
    "gs_event_timetable_monitor not present");

  /* create a solver */
  gfs = KheGroupByFrameSolverMake(tasking, max_days, resource_invariant,
    common_frame, etm, ts, a);

  /* do group by frame for each resource type */
  if( rt != NULL )
    KheGroupByFrameForResourceType(gfs, rt);
  else
    for( i = 0;  i < KheInstanceResourceTypeCount(ins);  i++ )
      KheGroupByFrameForResourceType(gfs, KheInstanceResourceType(ins, i));

  /* free the solver and return */
  res = gfs->soln_changed;
  KheSolnArenaEnd(soln, a);
  /* KheGroupByFrameSolverFree(gfs); */
  if( DEBUG1 )
    fprintf(stderr, "] KheTaskingGroupByFrame returning %s\n",
      res ? "true" : "false");
  return res;
}
